# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
from __future__ import annotations

import os
import re
import sys
import traceback
import zipfile
from collections.abc import Callable
from concurrent.futures import Future
from typing import Any

import aqt.deckchooser
import aqt.forms
import aqt.modelchooser
from anki import importing
from anki.importing.anki2 import MediaMapInvalid, V2ImportIntoV1
from anki.importing.apkg import AnkiPackageImporter
from aqt.import_export.importing import ColpkgImporter
from aqt.main import AnkiQt, gui_hooks
from aqt.qt import *
from aqt.utils import (
    HelpPage,
    disable_help_button,
    getFile,
    getText,
    openHelp,
    showInfo,
    showText,
    showWarning,
    tooltip,
    tr,
)


class ChangeMap(QDialog):
    def __init__(self, mw: AnkiQt, model: dict, current: str) -> None:
        QDialog.__init__(self, mw, Qt.WindowType.Window)
        self.mw = mw
        self.model = model
        self.frm = aqt.forms.changemap.Ui_ChangeMap()
        self.frm.setupUi(self)
        disable_help_button(self)
        n = 0
        setCurrent = False
        for field in self.model["flds"]:
            item = QListWidgetItem(tr.importing_map_to(val=field["name"]))
            self.frm.fields.addItem(item)
            if current == field["name"]:
                setCurrent = True
                self.frm.fields.setCurrentRow(n)
            n += 1
        self.frm.fields.addItem(QListWidgetItem(tr.importing_map_to_tags()))
        self.frm.fields.addItem(QListWidgetItem(tr.importing_ignore_field()))
        if not setCurrent:
            if current == "_tags":
                self.frm.fields.setCurrentRow(n)
            else:
                self.frm.fields.setCurrentRow(n + 1)
        self.field: str | None = None

    def getField(self) -> str | None:
        self.exec()
        return self.field

    def accept(self) -> None:
        row = self.frm.fields.currentRow()
        if row < len(self.model["flds"]):
            self.field = self.model["flds"][row]["name"]
        elif row == self.frm.fields.count() - 2:
            self.field = "_tags"
        else:
            self.field = None
        QDialog.accept(self)

    def reject(self) -> None:
        self.accept()


# called by importFile() when importing a mappable file like .csv
# ImportType = Union[Importer,AnkiPackageImporter, TextImporter]


class ImportDialog(QDialog):
    _DEFAULT_FILE_DELIMITER = "\t"

    def __init__(self, mw: AnkiQt, importer: Any) -> None:
        QDialog.__init__(self, mw, Qt.WindowType.Window)
        self.mw = mw
        self.importer = importer
        self.frm = aqt.forms.importing.Ui_ImportDialog()
        self.frm.setupUi(self)
        help_button = self.frm.buttonBox.button(QDialogButtonBox.StandardButton.Help)
        assert help_button is not None
        qconnect(
            help_button.clicked,
            self.helpRequested,
        )
        disable_help_button(self)
        self.setupMappingFrame()
        self.setupOptions()
        self.modelChanged()
        self.frm.autoDetect.setVisible(self.importer.needDelimiter)
        gui_hooks.current_note_type_did_change.append(self.modelChanged)
        qconnect(self.frm.autoDetect.clicked, self.onDelimiter)
        self.updateDelimiterButtonText()
        assert self.mw.pm.profile is not None
        self.frm.allowHTML.setChecked(self.mw.pm.profile.get("allowHTML", True))
        qconnect(self.frm.importMode.currentIndexChanged, self.importModeChanged)
        self.frm.importMode.setCurrentIndex(self.mw.pm.profile.get("importMode", 1))
        self.frm.tagModified.setText(self.mw.pm.profile.get("tagModified", ""))
        self.frm.tagModified.setCol(self.mw.col)
        # import button
        b = QPushButton(tr.actions_import())
        self.frm.buttonBox.addButton(b, QDialogButtonBox.ButtonRole.AcceptRole)
        self.exec()

    def setupOptions(self) -> None:
        self.model = self.mw.col.models.current()
        self.modelChooser = aqt.modelchooser.ModelChooser(
            self.mw, self.frm.modelArea, label=False
        )
        self.deck = aqt.deckchooser.DeckChooser(self.mw, self.frm.deckArea, label=False)

    def modelChanged(self, unused: Any | None = None) -> None:
        self.importer.model = self.mw.col.models.current()
        self.importer.initMapping()
        self.showMapping()

    def onDelimiter(self) -> None:
        # Open a modal dialog to enter an delimiter
        # Todo/Idea Constrain the maximum width, so it doesn't take up that much screen space
        delim, ok = getText(
            tr.importing_by_default_anki_will_detect_the(),
            self,
            help=HelpPage.IMPORTING,
        )

        # If the modal dialog has been confirmed, update the delimiter
        if ok:
            # Check if the entered value is valid and if not fallback to default
            # at the moment every single character entry as well as '\t' is valid

            delim = delim if len(delim) > 0 else self._DEFAULT_FILE_DELIMITER
            delim = delim.replace("\\t", "\t")  # un-escape it
            if len(delim) > 1:
                showWarning(
                    tr.importing_multicharacter_separators_are_not_supported_please()
                )
                return
            self.hideMapping()

            def updateDelim() -> None:
                self.importer.delimiter = delim
                self.importer.updateDelimiter()
                self.updateDelimiterButtonText()

            self.showMapping(hook=updateDelim)

        else:
            # If the operation has been canceled, do not do anything
            pass

    def updateDelimiterButtonText(self) -> None:
        if not self.importer.needDelimiter:
            return
        if self.importer.delimiter:
            d = self.importer.delimiter
        else:
            d = self.importer.dialect.delimiter
        if d == "\t":
            d = tr.importing_tab()
        elif d == ",":
            d = tr.importing_comma()
        elif d == " ":
            d = tr.studying_space()
        elif d == ";":
            d = tr.importing_semicolon()
        elif d == ":":
            d = tr.importing_colon()
        else:
            d = repr(d)
        txt = tr.importing_fields_separated_by(val=d)
        self.frm.autoDetect.setText(txt)

    def accept(self) -> None:
        self.importer.mapping = self.mapping
        if not self.importer.mappingOk():
            showWarning(tr.importing_the_first_field_of_the_note())
            return
        self.importer.importMode = self.frm.importMode.currentIndex()
        assert self.mw.pm.profile is not None
        self.mw.pm.profile["importMode"] = self.importer.importMode
        self.importer.allowHTML = self.frm.allowHTML.isChecked()
        self.mw.pm.profile["allowHTML"] = self.importer.allowHTML
        self.importer.tagModified = self.frm.tagModified.text()
        self.mw.pm.profile["tagModified"] = self.importer.tagModified
        self.mw.col.set_aux_notetype_config(
            self.importer.model["id"], "lastDeck", self.deck.selected_deck_id
        )
        self.mw.col.models.save(self.importer.model, updateReqs=False)
        self.mw.progress.start()

        def on_done(future: Future) -> None:
            self.mw.progress.finish()

            try:
                future.result()
            except UnicodeDecodeError:
                showUnicodeWarning()
                return
            except Exception as e:
                msg = f"{tr.importing_failed_debug_info()}\n"
                err = repr(str(e))
                if "1-character string" in err:
                    msg += err
                elif "invalidTempFolder" in err:
                    msg += self.mw.errorHandler.tempFolderMsg()
                else:
                    msg += traceback.format_exc()
                showText(msg)
                return
            else:
                txt = f"{tr.importing_importing_complete()}\n"
                if self.importer.log:
                    txt += "\n".join(self.importer.log)
                self.close()
                showText(txt, plain_text_edit=True)
                self.mw.reset()

        self.mw.taskman.run_in_background(self.importer.run, on_done)

    def setupMappingFrame(self) -> None:
        # qt seems to have a bug with adding/removing from a grid, so we add
        # to a separate object and add/remove that instead
        self.frame = QFrame(self.frm.mappingArea)
        self.frm.mappingArea.setWidget(self.frame)
        self.mapbox = QVBoxLayout(self.frame)
        self.mapbox.setContentsMargins(0, 0, 0, 0)
        self.mapwidget: QWidget | None = None

    def hideMapping(self) -> None:
        self.frm.mappingGroup.hide()

    def showMapping(
        self, keepMapping: bool = False, hook: Callable | None = None
    ) -> None:
        if hook:
            hook()
        if not keepMapping:
            self.mapping = self.importer.mapping
        self.frm.mappingGroup.show()
        assert self.importer.fields()
        # set up the mapping grid
        if self.mapwidget:
            self.mapbox.removeWidget(self.mapwidget)
            self.mapwidget.deleteLater()
        self.mapwidget = QWidget()
        self.mapbox.addWidget(self.mapwidget)
        self.grid = QGridLayout(self.mapwidget)
        self.mapwidget.setLayout(self.grid)
        self.grid.setContentsMargins(3, 3, 3, 3)
        self.grid.setSpacing(6)
        for num in range(len(self.mapping)):
            text = tr.importing_field_of_file_is(val=num + 1)
            self.grid.addWidget(QLabel(text), num, 0)
            if self.mapping[num] == "_tags":
                text = tr.importing_mapped_to_tags()
            elif self.mapping[num]:
                text = tr.importing_mapped_to(val=self.mapping[num])
            else:
                text = tr.importing_ignored()
            self.grid.addWidget(QLabel(text), num, 1)
            button = QPushButton(tr.importing_change())
            self.grid.addWidget(button, num, 2)
            qconnect(button.clicked, lambda _, s=self, n=num: s.changeMappingNum(n))

    def changeMappingNum(self, n: int) -> None:
        f = ChangeMap(self.mw, self.importer.model, self.mapping[n]).getField()
        try:
            # make sure we don't have it twice
            index = self.mapping.index(f)
            self.mapping[index] = None
        except ValueError:
            pass
        self.mapping[n] = f
        if getattr(self.importer, "delimiter", False):
            self.savedDelimiter = self.importer.delimiter

            def updateDelim() -> None:
                self.importer.delimiter = self.savedDelimiter

            self.showMapping(hook=updateDelim, keepMapping=True)
        else:
            self.showMapping(keepMapping=True)

    def reject(self) -> None:
        self.modelChooser.cleanup()
        self.deck.cleanup()
        gui_hooks.current_note_type_did_change.remove(self.modelChanged)
        QDialog.reject(self)

    def helpRequested(self) -> None:
        openHelp(HelpPage.IMPORTING)

    def importModeChanged(self, newImportMode: int) -> None:
        if newImportMode == 0:
            self.frm.tagModified.setEnabled(True)
        else:
            self.frm.tagModified.setEnabled(False)


def showUnicodeWarning() -> None:
    """Shorthand to show a standard warning."""
    showWarning(tr.importing_selected_file_was_not_in_utf8())


def onImport(mw: AnkiQt) -> None:
    filt = ";;".join([x[0] for x in importing.importers(mw.col)])
    file = getFile(mw, tr.actions_import(), None, key="import", filter=filt)
    if not file:
        return
    file = str(file)

    head, ext = os.path.splitext(file)
    ext = ext.lower()
    if ext == ".anki":
        showInfo(tr.importing_anki_files_are_from_a_very())
        return
    elif ext == ".anki2":
        showInfo(tr.importing_anki2_files_are_not_directly_importable())
        return

    importFile(mw, file)


def importFile(mw: AnkiQt, file: str) -> None:
    importerClass = None
    done = False
    for i in importing.importers(mw.col):
        if done:
            break
        for mext in re.findall(r"[( ]?\*\.(.+?)[) ]", i[0]):
            if file.endswith(f".{mext}"):
                importerClass = i[1]
                done = True
                break
    if not importerClass:
        # if no matches, assume TSV
        importerClass = importing.importers(mw.col)[0][1]
    importer = importerClass(mw.col, file)
    # need to show import dialog?
    if importer.needMapper:
        # make sure we can load the file first
        mw.progress.start(immediate=True)
        try:
            importer.open()
            mw.progress.finish()
            ImportDialog(mw, importer)
        except UnicodeDecodeError:
            mw.progress.finish()
            showUnicodeWarning()
            return
        except Exception as e:
            mw.progress.finish()
            msg = repr(str(e))
            if msg == "'unknownFormat'":
                showWarning(tr.importing_unknown_file_format())
            else:
                msg = f"{tr.importing_failed_debug_info()}\n"
                msg += str(traceback.format_exc())
                showText(msg)
            return
        finally:
            importer.close()
    else:
        # if it's an apkg/zip, first test it's a valid file
        if isinstance(importer, AnkiPackageImporter):
            # we need to ask whether to import/replace; if it's
            # a colpkg file then the rest of the import process
            # will happen in setupApkgImport()
            if not setupApkgImport(mw, importer):
                return

        # importing non-colpkg files
        mw.progress.start(immediate=True)

        def on_done(future: Future) -> None:
            mw.progress.finish()
            try:
                future.result()
            except zipfile.BadZipfile:
                showWarning(invalidZipMsg())
            except MediaMapInvalid:
                showWarning(
                    "Unable to read file. It probably requires a newer version of Anki to import."
                )
            except V2ImportIntoV1:
                showWarning(
                    """\
To import this deck, please click the Update button at the top of the deck list, then try again."""
                )
            except Exception as e:
                err = repr(str(e))
                if "invalidFile" in err:
                    msg = tr.importing_invalid_file_please_restore_from_backup()
                    showWarning(msg)
                elif "invalidTempFolder" in err:
                    showWarning(mw.errorHandler.tempFolderMsg())
                elif "readonly" in err:
                    showWarning(tr.importing_unable_to_import_from_a_readonly())
                else:
                    msg = f"{tr.importing_failed_debug_info()}\n"
                    traceback.print_exc(file=sys.stdout)
                    msg += str(e)
                    showText(msg)
            else:
                log = "\n".join(importer.log)
                if "\n" not in log:
                    tooltip(log)
                else:
                    showText(log, plain_text_edit=True)

            mw.reset()

        mw.taskman.run_in_background(importer.run, on_done)


def invalidZipMsg() -> str:
    return tr.importing_this_file_does_not_appear_to()


def setupApkgImport(mw: AnkiQt, importer: AnkiPackageImporter) -> bool:
    base = os.path.basename(importer.file).lower()
    full = (
        (base == "collection.apkg")
        or re.match("backup-.*\\.apkg", base)
        or base.endswith(".colpkg")
    )
    if not full:
        # adding
        return True
    ColpkgImporter.do_import(mw, importer.file)
    return False
    return False



# Form implementation generated from reading ui file 'qt/aqt/forms/addcards.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(750, 493)
        Dialog.setMinimumSize(QtCore.QSize(400, 400))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("icons:anki.png"), QtGui.QIcon.Mode.Normal, QtGui.QIcon.State.Off)
        Dialog.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(parent=Dialog)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_3.setContentsMargins(12, 6, 12, 12)
        self.verticalLayout_3.setSpacing(12)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(-1, -1, -1, 0)
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.modelArea = QtWidgets.QWidget(parent=self.centralwidget)
        self.modelArea.setMinimumSize(QtCore.QSize(0, 10))
        self.modelArea.setObjectName("modelArea")
        self.horizontalLayout.addWidget(self.modelArea)
        self.deckArea = QtWidgets.QWidget(parent=self.centralwidget)
        self.deckArea.setObjectName("deckArea")
        self.horizontalLayout.addWidget(self.deckArea)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.fieldsArea = QtWidgets.QWidget(parent=self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(10)
        sizePolicy.setHeightForWidth(self.fieldsArea.sizePolicy().hasHeightForWidth())
        self.fieldsArea.setSizePolicy(sizePolicy)
        self.fieldsArea.setAutoFillBackground(True)
        self.fieldsArea.setObjectName("fieldsArea")
        self.verticalLayout_3.addWidget(self.fieldsArea)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=self.centralwidget)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.NoButton)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_3.addWidget(self.buttonBox)
        Dialog.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(parent=Dialog)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 750, 22))
        self.menubar.setObjectName("menubar")
        self.menu_Edit = QtWidgets.QMenu(parent=self.menubar)
        self.menu_Edit.setObjectName("menu_Edit")
        Dialog.setMenuBar(self.menubar)
        self.menubar.addAction(self.menu_Edit.menuAction())

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        self.menu_Edit.setTitle(tr.qt_accel_edit())


# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

"""
Anki maintains a cache of used tags so it can quickly present a list of tags
for autocomplete and in the browser. For efficiency, deletions are not
tracked, so unused tags can only be removed from the list with a DB check.

This module manages the tag cache and tags for notes.
"""

from __future__ import annotations

import pprint
import re
from collections.abc import Collection, Sequence
from typing import Match

import anki
import anki.collection
from anki import tags_pb2
from anki._legacy import DeprecatedNamesMixin, deprecated
from anki.collection import OpChanges, OpChangesWithCount
from anki.decks import DeckId
from anki.notes import NoteId
from anki.utils import ids2str

# public exports
TagTreeNode = tags_pb2.TagTreeNode
CompleteTagRequest = tags_pb2.CompleteTagRequest
MARKED_TAG = "marked"


class TagManager(DeprecatedNamesMixin):
    def __init__(self, col: anki.collection.Collection) -> None:
        self.col = col.weakref()

    # legacy add-on code expects a List return type
    def all(self) -> list[str]:
        return list(self.col._backend.all_tags())

    def __repr__(self) -> str:
        dict_ = dict(self.__dict__)
        del dict_["col"]
        return f"{super().__repr__()} {pprint.pformat(dict_, width=300)}"

    def tree(self) -> TagTreeNode:
        return self.col._backend.tag_tree()

    # Registering and fetching tags
    #############################################################

    def clear_unused_tags(self) -> OpChangesWithCount:
        return self.col._backend.clear_unused_tags()

    def set_collapsed(self, tag: str, collapsed: bool) -> OpChanges:
        "Set browser expansion state for tag, registering the tag if missing."
        return self.col._backend.set_tag_collapsed(name=tag, collapsed=collapsed)

    # Bulk addition/removal from specific notes
    #############################################################

    def bulk_add(self, note_ids: Sequence[NoteId], tags: str) -> OpChangesWithCount:
        """Add space-separate tags to provided notes, returning changed count."""
        return self.col._backend.add_note_tags(note_ids=note_ids, tags=tags)

    def bulk_remove(self, note_ids: Sequence[NoteId], tags: str) -> OpChangesWithCount:
        return self.col._backend.remove_note_tags(note_ids=note_ids, tags=tags)

    # Find&replace
    #############################################################

    def find_and_replace(
        self,
        note_ids: Sequence[int],
        search: str,
        replacement: str,
        regex: bool,
        match_case: bool,
    ) -> OpChangesWithCount:
        """Replace instances of 'search' with 'replacement' in tags.
        Each tag is matched separately. If the replacement results in an empty string,
        the tag will be removed."""
        return self.col._backend.find_and_replace_tag(
            note_ids=note_ids,
            search=search,
            replacement=replacement,
            regex=regex,
            match_case=match_case,
        )

    # Bulk addition/removal based on tag
    #############################################################

    def rename(self, old: str, new: str) -> OpChangesWithCount:
        "Rename provided tag and its children, returning number of changed notes."
        return self.col._backend.rename_tags(current_prefix=old, new_prefix=new)

    def remove(self, space_separated_tags: str) -> OpChangesWithCount:
        "Remove the provided tag(s) and their children from notes and the tag list."
        return self.col._backend.remove_tags(val=space_separated_tags)

    def reparent(self, tags: Sequence[str], new_parent: str) -> OpChangesWithCount:
        """Change the parent of the provided tags.
        If new_parent is empty, tags will be reparented to the top-level."""
        return self.col._backend.reparent_tags(tags=tags, new_parent=new_parent)

    # String-based utilities
    ##########################################################################

    def split(self, tags: str) -> list[str]:
        "Parse a string and return a list of tags."
        return [t for t in tags.replace("\u3000", " ").split(" ") if t]

    def join(self, tags: list[str]) -> str:
        "Join tags into a single string, with leading and trailing spaces."
        if not tags:
            return ""
        return f" {' '.join(tags)} "

    def rem_from_str(self, deltags: str, tags: str) -> str:
        "Delete tags if they exist."

        def wildcard(pat: str, repl: str) -> Match:
            pat = re.escape(pat).replace("\\*", ".*")
            return re.match(f"^{pat}$", repl, re.IGNORECASE)

        current_tags = self.split(tags)
        for del_tag in self.split(deltags):
            # find tags, ignoring case
            remove = []
            for cur_tag in current_tags:
                if (del_tag.lower() == cur_tag.lower()) or wildcard(del_tag, cur_tag):
                    remove.append(cur_tag)
            # remove them
            for rem in remove:
                current_tags.remove(rem)
        return self.join(current_tags)

    # List-based utilities
    ##########################################################################

    @deprecated(info="no-op - tags are now canonified when note is saved")
    def canonify(self, tag_list: list[str]) -> list[str]:
        return tag_list

    def in_list(self, tag: str, tags: list[str]) -> bool:
        "True if TAG is in TAGS. Ignore case."
        return tag.lower() in [t.lower() for t in tags]

    # legacy
    ##########################################################################

    def _legacy_register_notes(self, nids: list[int] | None = None) -> None:
        self.clear_unused_tags()

    def register(
        self, tags: Collection[str], usn: int | None = None, clear: bool = False
    ) -> None:
        print("tags.register() is deprecated and no longer works")

    def _legacy_bulk_add(self, ids: list[NoteId], tags: str, add: bool = True) -> None:
        "Add tags in bulk. TAGS is space-separated."
        if add:
            self.bulk_add(ids, tags)
        else:
            self.bulk_remove(ids, tags)

    def _legacy_bulk_rem(self, ids: list[NoteId], tags: str) -> None:
        self._legacy_bulk_add(ids, tags, False)

    @deprecated(info="no longer used by Anki, and will be removed in the future")
    def by_deck(self, did: DeckId, children: bool = False) -> list[str]:
        basequery = "select n.tags from cards c, notes n WHERE c.nid = n.id"
        if not children:
            query = f"{basequery} AND c.did=?"
            res = self.col.db.list(query, did)
            return list(set(self.split(" ".join(res))))
        dids = [did]
        for name, id in self.col.decks.children(did):
            dids.append(id)
        query = f"{basequery} AND c.did IN {ids2str(dids)}"
        res = self.col.db.list(query)
        return list(set(self.split(" ".join(res))))


TagManager.register_deprecated_attributes(
    registerNotes=(TagManager._legacy_register_notes, TagManager.clear_unused_tags),
    bulkAdd=(TagManager._legacy_bulk_add, TagManager.bulk_add),
    bulkRem=(TagManager._legacy_bulk_rem, TagManager.bulk_remove),
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/image_occlusion.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/image_occlusion.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import collection_pb2 as anki_dot_collection__pb2
from anki import generic_pb2 as anki_dot_generic__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1a\x61nki/image_occlusion.proto\x12\x14\x61nki.image_occlusion\x1a\x15\x61nki/collection.proto\x1a\x12\x61nki/generic.proto\"+\n\x1bGetImageForOcclusionRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\":\n\x1cGetImageForOcclusionResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\t\"\x8d\x01\n\x1c\x41\x64\x64ImageOcclusionNoteRequest\x12\x12\n\nimage_path\x18\x01 \x01(\t\x12\x12\n\nocclusions\x18\x02 \x01(\t\x12\x0e\n\x06header\x18\x03 \x01(\t\x12\x12\n\nback_extra\x18\x04 \x01(\t\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12\x13\n\x0bnotetype_id\x18\x06 \x01(\x03\"/\n\x1cGetImageOcclusionNoteRequest\x12\x0f\n\x07note_id\x18\x01 \x01(\x03\"\xb3\x05\n\x1dGetImageOcclusionNoteResponse\x12V\n\x04note\x18\x01 \x01(\x0b\x32\x46.anki.image_occlusion.GetImageOcclusionNoteResponse.ImageOcclusionNoteH\x00\x12\x0f\n\x05\x65rror\x18\x02 \x01(\tH\x00\x1a\x35\n\x16ImageOcclusionProperty\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\x1a\x84\x01\n\x13ImageOcclusionShape\x12\r\n\x05shape\x18\x01 \x01(\t\x12^\n\nproperties\x18\x02 \x03(\x0b\x32J.anki.image_occlusion.GetImageOcclusionNoteResponse.ImageOcclusionProperty\x1az\n\x0eImageOcclusion\x12W\n\x06shapes\x18\x01 \x03(\x0b\x32G.anki.image_occlusion.GetImageOcclusionNoteResponse.ImageOcclusionShape\x12\x0f\n\x07ordinal\x18\x02 \x01(\r\x1a\xe5\x01\n\x12ImageOcclusionNote\x12\x12\n\nimage_data\x18\x01 \x01(\x0c\x12V\n\nocclusions\x18\x02 \x03(\x0b\x32\x42.anki.image_occlusion.GetImageOcclusionNoteResponse.ImageOcclusion\x12\x0e\n\x06header\x18\x03 \x01(\t\x12\x12\n\nback_extra\x18\x04 \x01(\t\x12\x0c\n\x04tags\x18\x05 \x03(\t\x12\x17\n\x0fimage_file_name\x18\x06 \x01(\t\x12\x18\n\x10occlude_inactive\x18\x07 \x01(\x08\x42\x07\n\x05value\"x\n\x1fUpdateImageOcclusionNoteRequest\x12\x0f\n\x07note_id\x18\x01 \x01(\x03\x12\x12\n\nocclusions\x18\x02 \x01(\t\x12\x0e\n\x06header\x18\x03 \x01(\t\x12\x12\n\nback_extra\x18\x04 \x01(\t\x12\x0c\n\x04tags\x18\x05 \x03(\t\"5\n\x1eGetImageOcclusionFieldsRequest\x12\x13\n\x0bnotetype_id\x18\x01 \x01(\x03\"c\n\x1fGetImageOcclusionFieldsResponse\x12@\n\x06\x66ields\x18\x01 \x01(\x0b\x32\x30.anki.image_occlusion.ImageOcclusionFieldIndexes\"c\n\x1aImageOcclusionFieldIndexes\x12\x12\n\nocclusions\x18\x01 \x01(\r\x12\r\n\x05image\x18\x02 \x01(\r\x12\x0e\n\x06header\x18\x03 \x01(\r\x12\x12\n\nback_extra\x18\x04 \x01(\r2\xc8\x05\n\x15ImageOcclusionService\x12}\n\x14GetImageForOcclusion\x12\x31.anki.image_occlusion.GetImageForOcclusionRequest\x1a\x32.anki.image_occlusion.GetImageForOcclusionResponse\x12\x80\x01\n\x15GetImageOcclusionNote\x12\x32.anki.image_occlusion.GetImageOcclusionNoteRequest\x1a\x33.anki.image_occlusion.GetImageOcclusionNoteResponse\x12\x86\x01\n\x17GetImageOcclusionFields\x12\x34.anki.image_occlusion.GetImageOcclusionFieldsRequest\x1a\x35.anki.image_occlusion.GetImageOcclusionFieldsResponse\x12L\n\x19\x41\x64\x64ImageOcclusionNotetype\x12\x13.anki.generic.Empty\x1a\x1a.anki.collection.OpChanges\x12g\n\x15\x41\x64\x64ImageOcclusionNote\x12\x32.anki.image_occlusion.AddImageOcclusionNoteRequest\x1a\x1a.anki.collection.OpChanges\x12m\n\x18UpdateImageOcclusionNote\x12\x35.anki.image_occlusion.UpdateImageOcclusionNoteRequest\x1a\x1a.anki.collection.OpChanges2\x1e\n\x1c\x42\x61\x63kendImageOcclusionServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.image_occlusion_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_GETIMAGEFOROCCLUSIONREQUEST']._serialized_start=95
  _globals['_GETIMAGEFOROCCLUSIONREQUEST']._serialized_end=138
  _globals['_GETIMAGEFOROCCLUSIONRESPONSE']._serialized_start=140
  _globals['_GETIMAGEFOROCCLUSIONRESPONSE']._serialized_end=198
  _globals['_ADDIMAGEOCCLUSIONNOTEREQUEST']._serialized_start=201
  _globals['_ADDIMAGEOCCLUSIONNOTEREQUEST']._serialized_end=342
  _globals['_GETIMAGEOCCLUSIONNOTEREQUEST']._serialized_start=344
  _globals['_GETIMAGEOCCLUSIONNOTEREQUEST']._serialized_end=391
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE']._serialized_start=394
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE']._serialized_end=1085
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE_IMAGEOCCLUSIONPROPERTY']._serialized_start=532
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE_IMAGEOCCLUSIONPROPERTY']._serialized_end=585
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE_IMAGEOCCLUSIONSHAPE']._serialized_start=588
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE_IMAGEOCCLUSIONSHAPE']._serialized_end=720
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE_IMAGEOCCLUSION']._serialized_start=722
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE_IMAGEOCCLUSION']._serialized_end=844
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE_IMAGEOCCLUSIONNOTE']._serialized_start=847
  _globals['_GETIMAGEOCCLUSIONNOTERESPONSE_IMAGEOCCLUSIONNOTE']._serialized_end=1076
  _globals['_UPDATEIMAGEOCCLUSIONNOTEREQUEST']._serialized_start=1087
  _globals['_UPDATEIMAGEOCCLUSIONNOTEREQUEST']._serialized_end=1207
  _globals['_GETIMAGEOCCLUSIONFIELDSREQUEST']._serialized_start=1209
  _globals['_GETIMAGEOCCLUSIONFIELDSREQUEST']._serialized_end=1262
  _globals['_GETIMAGEOCCLUSIONFIELDSRESPONSE']._serialized_start=1264
  _globals['_GETIMAGEOCCLUSIONFIELDSRESPONSE']._serialized_end=1363
  _globals['_IMAGEOCCLUSIONFIELDINDEXES']._serialized_start=1365
  _globals['_IMAGEOCCLUSIONFIELDINDEXES']._serialized_end=1464
  _globals['_IMAGEOCCLUSIONSERVICE']._serialized_start=1467
  _globals['_IMAGEOCCLUSIONSERVICE']._serialized_end=2179
  _globals['_BACKENDIMAGEOCCLUSIONSERVICE']._serialized_start=2181
  _globals['_BACKENDIMAGEOCCLUSIONSERVICE']._serialized_end=2211
# @@protoc_insertion_point(module_scope)



# Form implementation generated from reading ui file 'qt/aqt/forms/preferences.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Preferences(object):
    def setupUi(self, Preferences):
        Preferences.setObjectName("Preferences")
        Preferences.resize(636, 638)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(Preferences)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tabWidget = QtWidgets.QTabWidget(parent=Preferences)
        self.tabWidget.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.tabWidget.setObjectName("tabWidget")
        self.appearanceTab = QtWidgets.QWidget()
        self.appearanceTab.setObjectName("appearanceTab")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.appearanceTab)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.generalGroup = QtWidgets.QGroupBox(parent=self.appearanceTab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.generalGroup.sizePolicy().hasHeightForWidth())
        self.generalGroup.setSizePolicy(sizePolicy)
        self.generalGroup.setObjectName("generalGroup")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.generalGroup)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.label_9 = QtWidgets.QLabel(parent=self.generalGroup)
        self.label_9.setObjectName("label_9")
        self.gridLayout_5.addWidget(self.label_9, 0, 0, 1, 1)
        self.video_driver = QtWidgets.QComboBox(parent=self.generalGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.video_driver.sizePolicy().hasHeightForWidth())
        self.video_driver.setSizePolicy(sizePolicy)
        self.video_driver.setObjectName("video_driver")
        self.gridLayout_5.addWidget(self.video_driver, 1, 1, 1, 1)
        self.video_driver_label = QtWidgets.QLabel(parent=self.generalGroup)
        self.video_driver_label.setObjectName("video_driver_label")
        self.gridLayout_5.addWidget(self.video_driver_label, 1, 0, 1, 1)
        self.lang = QtWidgets.QComboBox(parent=self.generalGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lang.sizePolicy().hasHeightForWidth())
        self.lang.setSizePolicy(sizePolicy)
        self.lang.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.lang.setObjectName("lang")
        self.gridLayout_5.addWidget(self.lang, 0, 1, 1, 1)
        self.check_for_updates = QtWidgets.QCheckBox(parent=self.generalGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.check_for_updates.sizePolicy().hasHeightForWidth())
        self.check_for_updates.setSizePolicy(sizePolicy)
        self.check_for_updates.setObjectName("check_for_updates")
        self.gridLayout_5.addWidget(self.check_for_updates, 2, 0, 1, 1)
        self.verticalLayout_8.addWidget(self.generalGroup)
        self.uiGroup = QtWidgets.QGroupBox(parent=self.appearanceTab)
        self.uiGroup.setObjectName("uiGroup")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.uiGroup)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.theme = QtWidgets.QComboBox(parent=self.uiGroup)
        self.theme.setObjectName("theme")
        self.gridLayout_2.addWidget(self.theme, 0, 1, 1, 1)
        self.uiScale = QtWidgets.QSpinBox(parent=self.uiGroup)
        self.uiScale.setSuffix("%")
        self.uiScale.setMinimum(100)
        self.uiScale.setMaximum(200)
        self.uiScale.setSingleStep(5)
        self.uiScale.setObjectName("uiScale")
        self.gridLayout_2.addWidget(self.uiScale, 2, 1, 1, 1)
        self.styleComboBox = QtWidgets.QComboBox(parent=self.uiGroup)
        self.styleComboBox.setCurrentText("")
        self.styleComboBox.setObjectName("styleComboBox")
        self.gridLayout_2.addWidget(self.styleComboBox, 1, 1, 1, 1)
        self.styleLabel = QtWidgets.QLabel(parent=self.uiGroup)
        self.styleLabel.setObjectName("styleLabel")
        self.gridLayout_2.addWidget(self.styleLabel, 1, 0, 1, 1)
        self.themeLabel = QtWidgets.QLabel(parent=self.uiGroup)
        self.themeLabel.setObjectName("themeLabel")
        self.gridLayout_2.addWidget(self.themeLabel, 0, 0, 1, 1)
        self.uiSizeLabel = QtWidgets.QLabel(parent=self.uiGroup)
        self.uiSizeLabel.setObjectName("uiSizeLabel")
        self.gridLayout_2.addWidget(self.uiSizeLabel, 2, 0, 1, 1)
        self.resetWindowSizes = QtWidgets.QPushButton(parent=self.uiGroup)
        self.resetWindowSizes.setObjectName("resetWindowSizes")
        self.gridLayout_2.addWidget(self.resetWindowSizes, 3, 0, 1, 2)
        self.verticalLayout_8.addWidget(self.uiGroup)
        self.distractionsGroup = QtWidgets.QGroupBox(parent=self.appearanceTab)
        self.distractionsGroup.setObjectName("distractionsGroup")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.distractionsGroup)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.reduce_motion = QtWidgets.QCheckBox(parent=self.distractionsGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.reduce_motion.sizePolicy().hasHeightForWidth())
        self.reduce_motion.setSizePolicy(sizePolicy)
        self.reduce_motion.setObjectName("reduce_motion")
        self.gridLayout_3.addWidget(self.reduce_motion, 3, 0, 1, 1)
        self.hide_bottom_bar = QtWidgets.QCheckBox(parent=self.distractionsGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.hide_bottom_bar.sizePolicy().hasHeightForWidth())
        self.hide_bottom_bar.setSizePolicy(sizePolicy)
        self.hide_bottom_bar.setObjectName("hide_bottom_bar")
        self.gridLayout_3.addWidget(self.hide_bottom_bar, 2, 0, 1, 1)
        self.bottomBarComboBox = QtWidgets.QComboBox(parent=self.distractionsGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bottomBarComboBox.sizePolicy().hasHeightForWidth())
        self.bottomBarComboBox.setSizePolicy(sizePolicy)
        self.bottomBarComboBox.setObjectName("bottomBarComboBox")
        self.gridLayout_3.addWidget(self.bottomBarComboBox, 2, 1, 1, 1)
        self.hide_top_bar = QtWidgets.QCheckBox(parent=self.distractionsGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.hide_top_bar.sizePolicy().hasHeightForWidth())
        self.hide_top_bar.setSizePolicy(sizePolicy)
        self.hide_top_bar.setObjectName("hide_top_bar")
        self.gridLayout_3.addWidget(self.hide_top_bar, 0, 0, 1, 1)
        self.minimalist_mode = QtWidgets.QCheckBox(parent=self.distractionsGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.minimalist_mode.sizePolicy().hasHeightForWidth())
        self.minimalist_mode.setSizePolicy(sizePolicy)
        self.minimalist_mode.setObjectName("minimalist_mode")
        self.gridLayout_3.addWidget(self.minimalist_mode, 4, 0, 1, 1)
        self.topBarComboBox = QtWidgets.QComboBox(parent=self.distractionsGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.topBarComboBox.sizePolicy().hasHeightForWidth())
        self.topBarComboBox.setSizePolicy(sizePolicy)
        self.topBarComboBox.setObjectName("topBarComboBox")
        self.gridLayout_3.addWidget(self.topBarComboBox, 0, 1, 1, 1)
        self.verticalLayout_8.addWidget(self.distractionsGroup)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_8.addItem(spacerItem)
        self.tabWidget.addTab(self.appearanceTab, "")
        self.tab_3 = QtWidgets.QWidget()
        self.tab_3.setObjectName("tab_3")
        self.review_options_layout = QtWidgets.QVBoxLayout(self.tab_3)
        self.review_options_layout.setObjectName("review_options_layout")
        self.schedulerGroup = QtWidgets.QGroupBox(parent=self.tab_3)
        self.schedulerGroup.setObjectName("schedulerGroup")
        self.verticalLayout_17 = QtWidgets.QVBoxLayout(self.schedulerGroup)
        self.verticalLayout_17.setObjectName("verticalLayout_17")
        self.gridLayout_4 = QtWidgets.QGridLayout()
        self.gridLayout_4.setSpacing(12)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.lrnCutoff = QtWidgets.QSpinBox(parent=self.schedulerGroup)
        self.lrnCutoff.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.lrnCutoff.setMaximum(999)
        self.lrnCutoff.setObjectName("lrnCutoff")
        self.gridLayout_4.addWidget(self.lrnCutoff, 1, 1, 1, 1)
        self.label_30 = QtWidgets.QLabel(parent=self.schedulerGroup)
        self.label_30.setObjectName("label_30")
        self.gridLayout_4.addWidget(self.label_30, 2, 0, 1, 1)
        self.dayOffset = QtWidgets.QSpinBox(parent=self.schedulerGroup)
        self.dayOffset.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.dayOffset.setMaximum(23)
        self.dayOffset.setObjectName("dayOffset")
        self.gridLayout_4.addWidget(self.dayOffset, 0, 1, 1, 1)
        self.label_24 = QtWidgets.QLabel(parent=self.schedulerGroup)
        self.label_24.setObjectName("label_24")
        self.gridLayout_4.addWidget(self.label_24, 1, 0, 1, 1)
        self.label_23 = QtWidgets.QLabel(parent=self.schedulerGroup)
        self.label_23.setObjectName("label_23")
        self.gridLayout_4.addWidget(self.label_23, 0, 0, 1, 1)
        self.timeLimit = QtWidgets.QSpinBox(parent=self.schedulerGroup)
        self.timeLimit.setMaximum(9999)
        self.timeLimit.setObjectName("timeLimit")
        self.gridLayout_4.addWidget(self.timeLimit, 2, 1, 1, 1)
        self.verticalLayout_17.addLayout(self.gridLayout_4)
        self.review_options_layout.addWidget(self.schedulerGroup)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.reviewerGroup = QtWidgets.QGroupBox(parent=self.tab_3)
        self.reviewerGroup.setObjectName("reviewerGroup")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.reviewerGroup)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.showPlayButtons = QtWidgets.QCheckBox(parent=self.reviewerGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.showPlayButtons.sizePolicy().hasHeightForWidth())
        self.showPlayButtons.setSizePolicy(sizePolicy)
        self.showPlayButtons.setObjectName("showPlayButtons")
        self.verticalLayout_5.addWidget(self.showPlayButtons)
        self.interrupt_audio = QtWidgets.QCheckBox(parent=self.reviewerGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.interrupt_audio.sizePolicy().hasHeightForWidth())
        self.interrupt_audio.setSizePolicy(sizePolicy)
        self.interrupt_audio.setObjectName("interrupt_audio")
        self.verticalLayout_5.addWidget(self.interrupt_audio)
        self.showProgress = QtWidgets.QCheckBox(parent=self.reviewerGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.showProgress.sizePolicy().hasHeightForWidth())
        self.showProgress.setSizePolicy(sizePolicy)
        self.showProgress.setObjectName("showProgress")
        self.verticalLayout_5.addWidget(self.showProgress)
        self.showEstimates = QtWidgets.QCheckBox(parent=self.reviewerGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.showEstimates.sizePolicy().hasHeightForWidth())
        self.showEstimates.setSizePolicy(sizePolicy)
        self.showEstimates.setObjectName("showEstimates")
        self.verticalLayout_5.addWidget(self.showEstimates)
        self.spacebar_rates_card = QtWidgets.QCheckBox(parent=self.reviewerGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.spacebar_rates_card.sizePolicy().hasHeightForWidth())
        self.spacebar_rates_card.setSizePolicy(sizePolicy)
        self.spacebar_rates_card.setObjectName("spacebar_rates_card")
        self.verticalLayout_5.addWidget(self.spacebar_rates_card)
        self.render_latex = QtWidgets.QCheckBox(parent=self.reviewerGroup)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.render_latex.sizePolicy().hasHeightForWidth())
        self.render_latex.setSizePolicy(sizePolicy)
        self.render_latex.setObjectName("render_latex")
        self.verticalLayout_5.addWidget(self.render_latex)
        self.url_schemes = QtWidgets.QPushButton(parent=self.reviewerGroup)
        self.url_schemes.setObjectName("url_schemes")
        self.verticalLayout_5.addWidget(self.url_schemes)
        self.horizontalLayout_5.addWidget(self.reviewerGroup)
        self.preferences_answer_keys = QtWidgets.QGroupBox(parent=self.tab_3)
        self.preferences_answer_keys.setObjectName("preferences_answer_keys")
        self.horizontalLayout_5.addWidget(self.preferences_answer_keys)
        self.review_options_layout.addLayout(self.horizontalLayout_5)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.review_options_layout.addItem(spacerItem1)
        self.tabWidget.addTab(self.tab_3, "")
        self.tab_1 = QtWidgets.QWidget()
        self.tab_1.setObjectName("tab_1")
        self.verticalLayout_10 = QtWidgets.QVBoxLayout(self.tab_1)
        self.verticalLayout_10.setObjectName("verticalLayout_10")
        self.editorGroupBox = QtWidgets.QGroupBox(parent=self.tab_1)
        self.editorGroupBox.setObjectName("editorGroupBox")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.editorGroupBox)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.pastePNG = QtWidgets.QCheckBox(parent=self.editorGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pastePNG.sizePolicy().hasHeightForWidth())
        self.pastePNG.setSizePolicy(sizePolicy)
        self.pastePNG.setObjectName("pastePNG")
        self.verticalLayout_9.addWidget(self.pastePNG)
        self.paste_strips_formatting = QtWidgets.QCheckBox(parent=self.editorGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.paste_strips_formatting.sizePolicy().hasHeightForWidth())
        self.paste_strips_formatting.setSizePolicy(sizePolicy)
        self.paste_strips_formatting.setObjectName("paste_strips_formatting")
        self.verticalLayout_9.addWidget(self.paste_strips_formatting)
        spacerItem2 = QtWidgets.QSpacerItem(20, 10, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
        self.verticalLayout_9.addItem(spacerItem2)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.deck_label = QtWidgets.QLabel(parent=self.editorGroupBox)
        self.deck_label.setObjectName("deck_label")
        self.horizontalLayout_7.addWidget(self.deck_label)
        self.useCurrent = QtWidgets.QComboBox(parent=self.editorGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.MinimumExpanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.useCurrent.sizePolicy().hasHeightForWidth())
        self.useCurrent.setSizePolicy(sizePolicy)
        self.useCurrent.setObjectName("useCurrent")
        self.useCurrent.addItem("")
        self.useCurrent.addItem("")
        self.horizontalLayout_7.addWidget(self.useCurrent)
        self.verticalLayout_9.addLayout(self.horizontalLayout_7)
        self.verticalLayout_10.addWidget(self.editorGroupBox)
        self.browserGroupBox = QtWidgets.QGroupBox(parent=self.tab_1)
        self.browserGroupBox.setObjectName("browserGroupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.browserGroupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.search_text_label = QtWidgets.QLabel(parent=self.browserGroupBox)
        self.search_text_label.setObjectName("search_text_label")
        self.horizontalLayout_2.addWidget(self.search_text_label)
        self.default_search_text = QtWidgets.QLineEdit(parent=self.browserGroupBox)
        self.default_search_text.setObjectName("default_search_text")
        self.horizontalLayout_2.addWidget(self.default_search_text)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.ignore_accents_in_search = QtWidgets.QCheckBox(parent=self.browserGroupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.ignore_accents_in_search.sizePolicy().hasHeightForWidth())
        self.ignore_accents_in_search.setSizePolicy(sizePolicy)
        self.ignore_accents_in_search.setObjectName("ignore_accents_in_search")
        self.verticalLayout.addWidget(self.ignore_accents_in_search)
        self.verticalLayout_10.addWidget(self.browserGroupBox)
        spacerItem3 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_10.addItem(spacerItem3)
        self.tabWidget.addTab(self.tab_1, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.tab_2)
        self.verticalLayout_4.setContentsMargins(12, 12, 12, 12)
        self.verticalLayout_4.setSpacing(12)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.groupBox_5 = QtWidgets.QGroupBox(parent=self.tab_2)
        self.groupBox_5.setObjectName("groupBox_5")
        self.verticalLayout_12 = QtWidgets.QVBoxLayout(self.groupBox_5)
        self.verticalLayout_12.setObjectName("verticalLayout_12")
        self.syncMedia = QtWidgets.QCheckBox(parent=self.groupBox_5)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.syncMedia.sizePolicy().hasHeightForWidth())
        self.syncMedia.setSizePolicy(sizePolicy)
        self.syncMedia.setObjectName("syncMedia")
        self.verticalLayout_12.addWidget(self.syncMedia)
        self.syncOnProgramOpen = QtWidgets.QCheckBox(parent=self.groupBox_5)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.syncOnProgramOpen.sizePolicy().hasHeightForWidth())
        self.syncOnProgramOpen.setSizePolicy(sizePolicy)
        self.syncOnProgramOpen.setObjectName("syncOnProgramOpen")
        self.verticalLayout_12.addWidget(self.syncOnProgramOpen)
        self.autoSyncMedia = QtWidgets.QCheckBox(parent=self.groupBox_5)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.autoSyncMedia.sizePolicy().hasHeightForWidth())
        self.autoSyncMedia.setSizePolicy(sizePolicy)
        self.autoSyncMedia.setObjectName("autoSyncMedia")
        self.verticalLayout_12.addWidget(self.autoSyncMedia)
        self.fullSync = QtWidgets.QCheckBox(parent=self.groupBox_5)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.fullSync.sizePolicy().hasHeightForWidth())
        self.fullSync.setSizePolicy(sizePolicy)
        self.fullSync.setObjectName("fullSync")
        self.verticalLayout_12.addWidget(self.fullSync)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.label_2 = QtWidgets.QLabel(parent=self.groupBox_5)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_4.addWidget(self.label_2)
        self.network_timeout = QtWidgets.QSpinBox(parent=self.groupBox_5)
        self.network_timeout.setMinimum(30)
        self.network_timeout.setMaximum(99999)
        self.network_timeout.setObjectName("network_timeout")
        self.horizontalLayout_4.addWidget(self.network_timeout)
        self.verticalLayout_12.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem4)
        self.media_log = QtWidgets.QPushButton(parent=self.groupBox_5)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.media_log.sizePolicy().hasHeightForWidth())
        self.media_log.setSizePolicy(sizePolicy)
        self.media_log.setText("")
        self.media_log.setObjectName("media_log")
        self.horizontalLayout_3.addWidget(self.media_log)
        self.verticalLayout_12.addLayout(self.horizontalLayout_3)
        self.verticalLayout_4.addWidget(self.groupBox_5)
        self.groupBox_6 = QtWidgets.QGroupBox(parent=self.tab_2)
        self.groupBox_6.setObjectName("groupBox_6")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.groupBox_6)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.syncUser = QtWidgets.QLabel(parent=self.groupBox_6)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.syncUser.sizePolicy().hasHeightForWidth())
        self.syncUser.setSizePolicy(sizePolicy)
        self.syncUser.setText("")
        self.syncUser.setWordWrap(True)
        self.syncUser.setObjectName("syncUser")
        self.horizontalLayout.addWidget(self.syncUser)
        self.syncLogout = QtWidgets.QPushButton(parent=self.groupBox_6)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.syncLogout.sizePolicy().hasHeightForWidth())
        self.syncLogout.setSizePolicy(sizePolicy)
        self.syncLogout.setAutoDefault(False)
        self.syncLogout.setObjectName("syncLogout")
        self.horizontalLayout.addWidget(self.syncLogout)
        self.syncLogin = QtWidgets.QPushButton(parent=self.groupBox_6)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.syncLogin.sizePolicy().hasHeightForWidth())
        self.syncLogin.setSizePolicy(sizePolicy)
        self.syncLogin.setAutoDefault(False)
        self.syncLogin.setObjectName("syncLogin")
        self.horizontalLayout.addWidget(self.syncLogin)
        self.verticalLayout_4.addWidget(self.groupBox_6)
        spacerItem5 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_4.addItem(spacerItem5)
        spacerItem6 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_4.addItem(spacerItem6)
        self.gridLayout_21 = QtWidgets.QGridLayout()
        self.gridLayout_21.setObjectName("gridLayout_21")
        self.custom_sync_url = QtWidgets.QLineEdit(parent=self.tab_2)
        self.custom_sync_url.setObjectName("custom_sync_url")
        self.gridLayout_21.addWidget(self.custom_sync_url, 0, 1, 1, 1)
        self.label = QtWidgets.QLabel(parent=self.tab_2)
        self.label.setObjectName("label")
        self.gridLayout_21.addWidget(self.label, 0, 0, 1, 1)
        self.verticalLayout_4.addLayout(self.gridLayout_21)
        self.tabWidget.addTab(self.tab_2, "")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.tab)
        self.verticalLayout_3.setContentsMargins(12, 12, 12, 12)
        self.verticalLayout_3.setSpacing(12)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.groupBox_8 = QtWidgets.QGroupBox(parent=self.tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_8.sizePolicy().hasHeightForWidth())
        self.groupBox_8.setSizePolicy(sizePolicy)
        self.groupBox_8.setObjectName("groupBox_8")
        self.verticalLayout_15 = QtWidgets.QVBoxLayout(self.groupBox_8)
        self.verticalLayout_15.setObjectName("verticalLayout_15")
        self.backup_explanation = QtWidgets.QLabel(parent=self.groupBox_8)
        self.backup_explanation.setWordWrap(True)
        self.backup_explanation.setObjectName("backup_explanation")
        self.verticalLayout_15.addWidget(self.backup_explanation)
        spacerItem7 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
        self.verticalLayout_15.addItem(spacerItem7)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.minutes_between_backups = QtWidgets.QSpinBox(parent=self.groupBox_8)
        self.minutes_between_backups.setMinimum(5)
        self.minutes_between_backups.setMaximum(9999)
        self.minutes_between_backups.setObjectName("minutes_between_backups")
        self.gridLayout.addWidget(self.minutes_between_backups, 1, 2, 1, 1)
        self.monthly_backups = QtWidgets.QSpinBox(parent=self.groupBox_8)
        self.monthly_backups.setMaximum(9999)
        self.monthly_backups.setObjectName("monthly_backups")
        self.gridLayout.addWidget(self.monthly_backups, 5, 2, 1, 1)
        spacerItem8 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem8, 1, 3, 1, 1)
        self.label_3 = QtWidgets.QLabel(parent=self.groupBox_8)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 2, 0, 1, 1)
        self.weekly_backups = QtWidgets.QSpinBox(parent=self.groupBox_8)
        self.weekly_backups.setMaximum(9999)
        self.weekly_backups.setObjectName("weekly_backups")
        self.gridLayout.addWidget(self.weekly_backups, 4, 2, 1, 1)
        self.label_6 = QtWidgets.QLabel(parent=self.groupBox_8)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 5, 0, 1, 1)
        self.label_5 = QtWidgets.QLabel(parent=self.groupBox_8)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 4, 0, 1, 1)
        self.label_7 = QtWidgets.QLabel(parent=self.groupBox_8)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 1, 0, 1, 1)
        self.daily_backups = QtWidgets.QSpinBox(parent=self.groupBox_8)
        self.daily_backups.setMaximum(9999)
        self.daily_backups.setObjectName("daily_backups")
        self.gridLayout.addWidget(self.daily_backups, 2, 2, 1, 1)
        spacerItem9 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem9, 1, 1, 1, 1)
        self.verticalLayout_15.addLayout(self.gridLayout)
        spacerItem10 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Fixed)
        self.verticalLayout_15.addItem(spacerItem10)
        self.openBackupFolder = QtWidgets.QLabel(parent=self.groupBox_8)
        self.openBackupFolder.setObjectName("openBackupFolder")
        self.verticalLayout_15.addWidget(self.openBackupFolder)
        self.verticalLayout_3.addWidget(self.groupBox_8)
        self.groupBox_9 = QtWidgets.QGroupBox(parent=self.tab)
        self.groupBox_9.setObjectName("groupBox_9")
        self.verticalLayout_13 = QtWidgets.QVBoxLayout(self.groupBox_9)
        self.verticalLayout_13.setObjectName("verticalLayout_13")
        self.label_4 = QtWidgets.QLabel(parent=self.groupBox_9)
        self.label_4.setWordWrap(True)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_13.addWidget(self.label_4)
        self.verticalLayout_3.addWidget(self.groupBox_9)
        spacerItem11 = QtWidgets.QSpacerItem(20, 59, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_3.addItem(spacerItem11)
        self.tabWidget.addTab(self.tab, "")
        self.tab_31 = QtWidgets.QWidget()
        self.tab_31.setObjectName("tab_31")
        self.verticalLayout_41 = QtWidgets.QVBoxLayout(self.tab_31)
        self.verticalLayout_41.setContentsMargins(12, 12, 12, 12)
        self.verticalLayout_41.setSpacing(12)
        self.verticalLayout_41.setObjectName("verticalLayout_41")
        self.label1 = QtWidgets.QLabel(parent=self.tab_31)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label1.sizePolicy().hasHeightForWidth())
        self.label1.setSizePolicy(sizePolicy)
        self.label1.setObjectName("label1")
        self.verticalLayout_41.addWidget(self.label1)
        spacerItem12 = QtWidgets.QSpacerItem(20, 12, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Maximum)
        self.verticalLayout_41.addItem(spacerItem12)
        self.groupBox_61 = QtWidgets.QGroupBox(parent=self.tab_31)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_61.sizePolicy().hasHeightForWidth())
        self.groupBox_61.setSizePolicy(sizePolicy)
        self.groupBox_61.setTitle("AnkiHub")
        self.groupBox_61.setObjectName("groupBox_61")
        self.horizontalLayout1 = QtWidgets.QHBoxLayout(self.groupBox_61)
        self.horizontalLayout1.setObjectName("horizontalLayout1")
        self.syncAnkiHubUser = QtWidgets.QLabel(parent=self.groupBox_61)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.syncAnkiHubUser.sizePolicy().hasHeightForWidth())
        self.syncAnkiHubUser.setSizePolicy(sizePolicy)
        self.syncAnkiHubUser.setText("")
        self.syncAnkiHubUser.setWordWrap(True)
        self.syncAnkiHubUser.setObjectName("syncAnkiHubUser")
        self.horizontalLayout1.addWidget(self.syncAnkiHubUser)
        self.syncAnkiHubLogout = QtWidgets.QPushButton(parent=self.groupBox_61)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.syncAnkiHubLogout.sizePolicy().hasHeightForWidth())
        self.syncAnkiHubLogout.setSizePolicy(sizePolicy)
        self.syncAnkiHubLogout.setAutoDefault(False)
        self.syncAnkiHubLogout.setObjectName("syncAnkiHubLogout")
        self.horizontalLayout1.addWidget(self.syncAnkiHubLogout)
        self.syncAnkiHubLogin = QtWidgets.QPushButton(parent=self.groupBox_61)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.syncAnkiHubLogin.sizePolicy().hasHeightForWidth())
        self.syncAnkiHubLogin.setSizePolicy(sizePolicy)
        self.syncAnkiHubLogin.setAutoDefault(False)
        self.syncAnkiHubLogin.setObjectName("syncAnkiHubLogin")
        self.horizontalLayout1.addWidget(self.syncAnkiHubLogin)
        self.verticalLayout_41.addWidget(self.groupBox_61)
        spacerItem13 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_41.addItem(spacerItem13)
        self.tabWidget.addTab(self.tab_31, "")
        self.verticalLayout_2.addWidget(self.tabWidget)
        self.someSettingsLabel = QtWidgets.QLabel(parent=Preferences)
        self.someSettingsLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.someSettingsLabel.setObjectName("someSettingsLabel")
        self.verticalLayout_2.addWidget(self.someSettingsLabel)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Preferences)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Close|QtWidgets.QDialogButtonBox.StandardButton.Help)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_2.addWidget(self.buttonBox)
        self.label_9.setBuddy(self.lang)
        self.video_driver_label.setBuddy(self.video_driver)
        self.styleLabel.setBuddy(self.styleComboBox)
        self.themeLabel.setBuddy(self.theme)
        self.uiSizeLabel.setBuddy(self.uiScale)
        self.label_30.setBuddy(self.timeLimit)
        self.label_24.setBuddy(self.lrnCutoff)
        self.label_23.setBuddy(self.dayOffset)
        self.deck_label.setBuddy(self.useCurrent)
        self.search_text_label.setBuddy(self.default_search_text)
        self.label_2.setBuddy(self.network_timeout)
        self.label.setBuddy(self.custom_sync_url)
        self.label_3.setBuddy(self.daily_backups)
        self.label_6.setBuddy(self.monthly_backups)
        self.label_5.setBuddy(self.weekly_backups)
        self.label_7.setBuddy(self.minutes_between_backups)

        self.retranslateUi(Preferences)
        self.tabWidget.setCurrentIndex(0)
        self.buttonBox.accepted.connect(Preferences.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Preferences.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Preferences)
        Preferences.setTabOrder(self.lang, self.video_driver)
        Preferences.setTabOrder(self.video_driver, self.check_for_updates)
        Preferences.setTabOrder(self.check_for_updates, self.theme)
        Preferences.setTabOrder(self.theme, self.styleComboBox)
        Preferences.setTabOrder(self.styleComboBox, self.uiScale)
        Preferences.setTabOrder(self.uiScale, self.resetWindowSizes)
        Preferences.setTabOrder(self.resetWindowSizes, self.hide_top_bar)
        Preferences.setTabOrder(self.hide_top_bar, self.topBarComboBox)
        Preferences.setTabOrder(self.topBarComboBox, self.hide_bottom_bar)
        Preferences.setTabOrder(self.hide_bottom_bar, self.bottomBarComboBox)
        Preferences.setTabOrder(self.bottomBarComboBox, self.reduce_motion)
        Preferences.setTabOrder(self.reduce_motion, self.minimalist_mode)
        Preferences.setTabOrder(self.minimalist_mode, self.dayOffset)
        Preferences.setTabOrder(self.dayOffset, self.lrnCutoff)
        Preferences.setTabOrder(self.lrnCutoff, self.timeLimit)
        Preferences.setTabOrder(self.timeLimit, self.showPlayButtons)
        Preferences.setTabOrder(self.showPlayButtons, self.interrupt_audio)
        Preferences.setTabOrder(self.interrupt_audio, self.showProgress)
        Preferences.setTabOrder(self.showProgress, self.showEstimates)
        Preferences.setTabOrder(self.showEstimates, self.spacebar_rates_card)
        Preferences.setTabOrder(self.spacebar_rates_card, self.render_latex)
        Preferences.setTabOrder(self.render_latex, self.url_schemes)
        Preferences.setTabOrder(self.url_schemes, self.pastePNG)
        Preferences.setTabOrder(self.pastePNG, self.paste_strips_formatting)
        Preferences.setTabOrder(self.paste_strips_formatting, self.useCurrent)
        Preferences.setTabOrder(self.useCurrent, self.default_search_text)
        Preferences.setTabOrder(self.default_search_text, self.ignore_accents_in_search)
        Preferences.setTabOrder(self.ignore_accents_in_search, self.syncMedia)
        Preferences.setTabOrder(self.syncMedia, self.syncOnProgramOpen)
        Preferences.setTabOrder(self.syncOnProgramOpen, self.autoSyncMedia)
        Preferences.setTabOrder(self.autoSyncMedia, self.fullSync)
        Preferences.setTabOrder(self.fullSync, self.network_timeout)
        Preferences.setTabOrder(self.network_timeout, self.media_log)
        Preferences.setTabOrder(self.media_log, self.syncLogout)
        Preferences.setTabOrder(self.syncLogout, self.syncLogin)
        Preferences.setTabOrder(self.syncLogin, self.custom_sync_url)
        Preferences.setTabOrder(self.custom_sync_url, self.minutes_between_backups)
        Preferences.setTabOrder(self.minutes_between_backups, self.daily_backups)
        Preferences.setTabOrder(self.daily_backups, self.weekly_backups)
        Preferences.setTabOrder(self.weekly_backups, self.monthly_backups)
        Preferences.setTabOrder(self.monthly_backups, self.syncAnkiHubLogout)
        Preferences.setTabOrder(self.syncAnkiHubLogout, self.syncAnkiHubLogin)
        Preferences.setTabOrder(self.syncAnkiHubLogin, self.buttonBox)
        Preferences.setTabOrder(self.buttonBox, self.tabWidget)

    def retranslateUi(self, Preferences):
        _translate = QtCore.QCoreApplication.translate
        Preferences.setWindowTitle(tr.preferences_preferences())
        self.generalGroup.setTitle(tr.preferences_general())
        self.label_9.setText(tr.preferences_language())
        self.video_driver_label.setText(tr.preferences_video_driver())
        self.check_for_updates.setText(tr.preferences_check_for_updates())
        self.uiGroup.setTitle(tr.preferences_user_interface())
        self.styleLabel.setText(tr.preferences_style())
        self.themeLabel.setText(tr.preferences_theme())
        self.uiSizeLabel.setText(tr.preferences_user_interface_size())
        self.resetWindowSizes.setText(tr.preferences_reset_window_sizes())
        self.distractionsGroup.setTitle(tr.preferences_distractions())
        self.reduce_motion.setToolTip(tr.preferences_reduce_motion_tooltip())
        self.reduce_motion.setText(tr.preferences_reduce_motion())
        self.hide_bottom_bar.setText(tr.preferences_hide_bottom_bar_during_review())
        self.hide_top_bar.setText(tr.preferences_hide_top_bar_during_review())
        self.minimalist_mode.setToolTip(tr.preferences_minimalist_mode_tooltip())
        self.minimalist_mode.setText(tr.preferences_minimalist_mode())
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.appearanceTab), tr.preferences_appearance())
        self.schedulerGroup.setTitle(tr.preferences_scheduler())
        self.lrnCutoff.setSuffix(tr.preferences_mins())
        self.label_30.setText(tr.preferences_timebox_time_limit())
        self.dayOffset.setSuffix(tr.preferences_hours_past_midnight())
        self.label_24.setText(tr.preferences_learn_ahead_limit())
        self.label_23.setText(tr.preferences_next_day_starts_at())
        self.timeLimit.setSuffix(tr.preferences_mins())
        self.reviewerGroup.setTitle(tr.preferences_review())
        self.showPlayButtons.setText(tr.preferences_show_play_buttons_on_cards_with())
        self.interrupt_audio.setText(tr.preferences_interrupt_current_audio_when_answering())
        self.showProgress.setText(tr.preferences_show_remaining_card_count())
        self.showEstimates.setText(tr.preferences_show_next_review_time_above_answer())
        self.spacebar_rates_card.setText(tr.preferences_spacebar_rates_card())
        self.render_latex.setText(tr.preferences_generate_latex_images_automatically())
        self.url_schemes.setText(tr.preferences_url_schemes())
        self.preferences_answer_keys.setTitle(tr.preferences_answer_keys())
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_3), tr.preferences_review())
        self.editorGroupBox.setTitle(tr.preferences_editing())
        self.pastePNG.setText(tr.preferences_paste_clipboard_images_as_png())
        self.paste_strips_formatting.setText(tr.preferences_paste_without_shift_key_strips_formatting())
        self.deck_label.setText(tr.preferences_default_deck())
        self.useCurrent.setItemText(0, tr.preferences_when_adding_default_to_current_deck())
        self.useCurrent.setItemText(1, tr.preferences_change_deck_depending_on_note_type())
        self.browserGroupBox.setTitle(tr.preferences_browsing())
        self.search_text_label.setText(tr.preferences_default_search_text())
        self.default_search_text.setPlaceholderText(tr.preferences_default_search_text_example())
        self.ignore_accents_in_search.setText(tr.preferences_ignore_accents_in_search())
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_1), tr.preferences_editing())
        self.groupBox_5.setTitle(tr.preferences_tab_synchronisation())
        self.syncMedia.setText(tr.preferences_synchronize_audio_and_images_too())
        self.syncOnProgramOpen.setText(tr.preferences_automatically_sync_on_profile_openclose())
        self.autoSyncMedia.setText(tr.preferences_periodically_sync_media())
        self.fullSync.setText(tr.preferences_on_next_sync_force_changes_in())
        self.label_2.setText(tr.preferences_network_timeout())
        self.network_timeout.setSuffix(tr.scheduling_seconds())
        self.groupBox_6.setTitle(tr.preferences_account())
        self.syncLogout.setText(tr.sync_log_out_button())
        self.syncLogin.setText(tr.sync_log_in_button())
        self.custom_sync_url.setPlaceholderText(tr.preferences_custom_sync_url_disclaimer())
        self.label.setText(tr.preferences_custom_sync_url())
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), tr.preferences_network())
        self.groupBox_8.setTitle(tr.preferences_backups())
        self.backup_explanation.setText(tr.preferences_backup_explanation())
        self.label_3.setText(tr.preferences_daily_backups())
        self.label_6.setText(tr.preferences_monthly_backups())
        self.label_5.setText(tr.preferences_weekly_backups())
        self.label_7.setText(tr.preferences_minutes_between_backups())
        self.openBackupFolder.setText(tr.preferences_you_can_restore_backups_via_fileswitch())
        self.groupBox_9.setTitle(tr.preferences_note())
        self.label_4.setText(tr.preferences_media_is_not_backed_up())
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), tr.preferences_backups())
        self.label1.setText(tr.preferences_third_party_description())
        self.syncAnkiHubLogout.setText(tr.sync_log_out_button())
        self.syncAnkiHubLogin.setText(tr.sync_log_in_button())
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_31), tr.preferences_third_party_services())
        self.someSettingsLabel.setText(tr.preferences_some_settings_will_take_effect_after())


# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


from __future__ import annotations

import json
import random
import time
from collections.abc import Sequence
from typing import Any

import anki.cards
import anki.collection
from anki.consts import *
from anki.lang import FormatTimeSpan
from anki.utils import base62, ids2str

# Card stats
##########################################################################

_legacy_nightmode = False


def _legacy_card_stats(
    col: anki.collection.Collection, card_id: anki.cards.CardId, include_revlog: bool
) -> str:
    "A quick hack to preserve compatibility with the old HTML string API."
    random_id = f"cardinfo-{base62(random.randint(0, 2**64 - 1))}"
    varName = random_id.replace("-", "")
    return f"""
<div id="{random_id}"></div>
<script src="js/vendor/bootstrap.bundle.min.js"></script>
<link href="pages/card-info-base.css" rel="stylesheet" />
<link href="pages/card-info.css" rel="stylesheet" />
<script src="pages/card-info.js"></script>
<script>
    if ({1 if _legacy_nightmode else 0}) {{
        document.documentElement.className = "night-mode";
    }}
    const {varName} = anki.setupCardInfo(document.getElementById('{random_id}'));
    {varName}.then((c) => c.$set({{ cardId: {card_id}, includeRevlog: {str(include_revlog).lower()} }}));
</script>
    """


class CardStats:
    """Do not use - this class is only left around for backwards compatibility."""

    def __init__(self, col: anki.collection.Collection, card: anki.cards.Card) -> None:
        if col:
            self.col = col.weakref()
        self.card = card
        self.txt = ""

    def report(self, include_revlog: bool = False) -> str:
        return _legacy_card_stats(self.col, self.card.id, include_revlog)

    # legacy

    def addLine(self, k: str, v: int | str) -> None:
        self.txt += self.makeLine(k, v)

    def makeLine(self, k: str, v: str | int) -> str:
        txt = "<tr><td align=start style='padding-right: 3px;'>"
        txt += f"<b>{k}</b></td><td>{v}</td></tr>"
        return txt

    def date(self, tm: float) -> str:
        return time.strftime("%Y-%m-%d", time.localtime(tm))

    def time(self, tm: float) -> str:
        return self.col.format_timespan(tm, context=FormatTimeSpan.PRECISE)


# Collection stats
##########################################################################

PERIOD_MONTH = 0
PERIOD_YEAR = 1
PERIOD_LIFE = 2

colYoung = "#7c7"
colMature = "#070"
colCum = "rgba(0,0,0,0.9)"
colLearn = "#00F"
colRelearn = "#c00"
colCram = "#ff0"
colIvl = "#077"
colHour = "#ccc"
colTime = "#770"
colUnseen = "#000"
colSusp = "#ff0"


class CollectionStats:
    def __init__(self, col: anki.collection.Collection) -> None:
        self.col = col.weakref()
        self._stats = None
        self.type = PERIOD_MONTH
        self.width = 600
        self.height = 200
        self.wholeCollection = False

    # assumes jquery & plot are available in document
    def report(self, type: int = PERIOD_MONTH) -> str:
        # 0=month, 1=year, 2=deck life
        self.type = type
        from .statsbg import bg

        txt = self.css % bg
        txt += self._section(self.todayStats())
        txt += self._section(self.dueGraph())
        txt += self.repsGraphs()
        txt += self._section(self.introductionGraph())
        txt += self._section(self.ivlGraph())
        txt += self._section(self.hourGraph())
        txt += self._section(self.easeGraph())
        txt += self._section(self.cardGraph())
        txt += self._section(self.footer())
        return "<center>%s</center>" % txt

    def _section(self, txt: str) -> str:
        return "<div class=section>%s</div>" % txt

    css = """
<style>
h1 { margin-bottom: 0; margin-top: 1em; }
.pielabel { text-align:center; padding:0px; color:white; }
body:not(.night_mode) {background-image: url(data:image/png;base64,%s); }
@media print {
    .section { page-break-inside: avoid; padding-top: 5mm; }
}
body { direction: ltr !important; }
</style>
"""

    # Today stats
    ######################################################################

    def todayStats(self) -> str:
        b = self._title("Today")
        # studied today
        lim = self._revlogLimit()
        if lim:
            lim = " and " + lim
        cards, thetime, failed, lrn, rev, relrn, filt = self.col.db.first(
            f"""
select count(), sum(time)/1000,
sum(case when ease = 1 then 1 else 0 end), /* failed */
sum(case when type = {REVLOG_LRN} then 1 else 0 end), /* learning */
sum(case when type = {REVLOG_REV} then 1 else 0 end), /* review */
sum(case when type = {REVLOG_RELRN} then 1 else 0 end), /* relearn */
sum(case when type = {REVLOG_CRAM} then 1 else 0 end) /* filter */
from revlog where type != {REVLOG_RESCHED} and id > ? """
            + lim,
            (self.col.sched.day_cutoff - 86400) * 1000,
        )
        cards = cards or 0
        thetime = thetime or 0
        failed = failed or 0
        lrn = lrn or 0
        rev = rev or 0
        relrn = relrn or 0
        filt = filt or 0

        # studied
        def bold(s: str) -> str:
            return "<b>" + str(s) + "</b>"

        if cards:
            b += self.col._backend.studied_today_message(
                cards=cards, seconds=float(thetime)
            )
            # again/pass count
            b += "<br>" + "Again count: %s" % bold(str(failed))
            if cards:
                b += " " + "(%s correct)" % bold(
                    "%0.1f%%" % ((1 - failed / float(cards)) * 100)
                )
            # type breakdown
            b += "<br>"
            b += "Learn: %(a)s, Review: %(b)s, Relearn: %(c)s, Filtered: %(d)s" % dict(
                a=bold(str(lrn)),
                b=bold(str(rev)),
                c=bold(str(relrn)),
                d=bold(str(filt)),
            )
            # mature today
            mcnt, msum = self.col.db.first(
                """
    select count(), sum(case when ease = 1 then 0 else 1 end) from revlog
    where lastIvl >= 21 and id > ?"""
                + lim,
                (self.col.sched.day_cutoff - 86400) * 1000,
            )
            b += "<br>"
            if mcnt:
                b += "Correct answers on mature cards: %(a)d/%(b)d (%(c).1f%%)" % dict(
                    a=msum, b=mcnt, c=(msum / float(mcnt) * 100)
                )
            else:
                b += "No mature cards were studied today."
        else:
            b += "No cards have been studied today."
        return b

    # Due and cumulative due
    ######################################################################

    def get_start_end_chunk(self, by: str = "review") -> tuple[int, int | None, int]:
        start = 0
        if self.type == PERIOD_MONTH:
            end, chunk = 31, 1
        elif self.type == PERIOD_YEAR:
            end, chunk = 52, 7
        else:  #  self.type == 2:
            end = None
            if self._deckAge(by) <= 100:
                chunk = 1
            elif self._deckAge(by) <= 700:
                chunk = 7
            else:
                chunk = 31
        return start, end, chunk

    def dueGraph(self) -> str:
        start, end, chunk = self.get_start_end_chunk()
        d = self._due(start, end, chunk)
        yng = []
        mtr = []
        tot = 0
        totd = []
        for day in d:
            yng.append((day[0], day[1]))
            mtr.append((day[0], day[2]))
            tot += day[1] + day[2]
            totd.append((day[0], tot))
        data: Any = [
            dict(data=mtr, color=colMature, label="Mature"),
            dict(data=yng, color=colYoung, label="Young"),
        ]
        if len(totd) > 1:
            data.append(
                dict(
                    data=totd,
                    color=colCum,
                    label="Cumulative",
                    yaxis=2,
                    bars={"show": False},
                    lines=dict(show=True),
                    stack=False,
                )
            )
        txt = self._title("Forecast", "The number of reviews due in the future.")
        xaxis = dict(tickDecimals=0, min=-0.5)
        if end is not None:
            xaxis["max"] = end - 0.5
        txt += self._graph(
            id="due",
            data=data,
            xunit=chunk,
            ylabel2="Cumulative Cards",
            conf=dict(
                xaxis=xaxis,
                yaxes=[dict(min=0), dict(min=0, tickDecimals=0, position="right")],
            ),
        )
        txt += self._dueInfo(tot, len(totd) * chunk)
        return txt

    def _dueInfo(self, tot: int, num: int) -> str:
        i: list[str] = []
        self._line(
            i,
            "Total",
            self.col.tr.statistics_reviews(reviews=tot),
        )
        self._line(i, "Average", self._avgDay(tot, num, "reviews"))
        tomorrow = self.col.db.scalar(
            f"""
select count() from cards where did in %s and queue in ({QUEUE_TYPE_REV},{QUEUE_TYPE_DAY_LEARN_RELEARN})
and due = ?"""
            % self._limit(),
            self.col.sched.today + 1,
        )
        tomorrow = "%d cards" % tomorrow
        self._line(i, "Due tomorrow", tomorrow)
        return self._lineTbl(i)

    def _due(
        self, start: int | None = None, end: int | None = None, chunk: int = 1
    ) -> Any:
        lim = ""
        if start is not None:
            lim += " and due-%d >= %d" % (self.col.sched.today, start)
        if end is not None:
            lim += " and day < %d" % end
        return self.col.db.all(
            f"""
select (due-?)/? as day,
sum(case when ivl < 21 then 1 else 0 end), -- yng
sum(case when ivl >= 21 then 1 else 0 end) -- mtr
from cards
where did in %s and queue in ({QUEUE_TYPE_REV},{QUEUE_TYPE_DAY_LEARN_RELEARN})
%s
group by day order by day"""
            % (self._limit(), lim),
            self.col.sched.today,
            chunk,
        )

    # Added, reps and time spent
    ######################################################################

    def introductionGraph(self) -> str:
        start, days, chunk = self.get_start_end_chunk()
        data = self._added(days, chunk)
        if not data:
            return ""
        conf: dict[str, Any] = dict(
            xaxis=dict(tickDecimals=0, max=0.5),
            yaxes=[dict(min=0), dict(position="right", min=0)],
        )
        if days is not None:
            conf["xaxis"]["min"] = -days + 0.5

        def plot(id: str, data: Any, ylabel: str, ylabel2: str) -> str:
            return self._graph(
                id, data=data, conf=conf, xunit=chunk, ylabel=ylabel, ylabel2=ylabel2
            )

        # graph
        repdata, repsum = self._splitRepData(data, ((1, colLearn, ""),))
        txt = self._title("Added", "The number of new cards you have added.")
        txt += plot("intro", repdata, ylabel="Cards", ylabel2="Cumulative Cards")
        # total and per day average
        tot = sum(i[1] for i in data)
        period = self._periodDays()
        if not period:
            # base off date of earliest added card
            period = self._deckAge("add")
        i: list[str] = []
        self._line(i, "Total", "%d cards" % tot)
        self._line(i, "Average", self._avgDay(tot, period, "cards"))
        txt += self._lineTbl(i)

        return txt

    def repsGraphs(self) -> str:
        start, days, chunk = self.get_start_end_chunk()
        data = self._done(days, chunk)
        if not data:
            return ""
        conf: dict[str, Any] = dict(
            xaxis=dict(tickDecimals=0, max=0.5),
            yaxes=[dict(min=0), dict(position="right", min=0)],
        )
        if days is not None:
            conf["xaxis"]["min"] = -days + 0.5

        def plot(id: str, data: Any, ylabel: str, ylabel2: str) -> str:
            return self._graph(
                id, data=data, conf=conf, xunit=chunk, ylabel=ylabel, ylabel2=ylabel2
            )

        # reps
        (repdata, repsum) = self._splitRepData(
            data,
            (
                (3, colMature, "Mature"),
                (2, colYoung, "Young"),
                (4, colRelearn, "Relearn"),
                (1, colLearn, "Learn"),
                (5, colCram, "Cram"),
            ),
        )
        txt1 = self._title("Review Count", "The number of questions you have answered.")
        txt1 += plot("reps", repdata, ylabel="Answers", ylabel2="Cumulative Answers")
        (daysStud, fstDay) = self._daysStudied()
        rep, tot = self._ansInfo(repsum, daysStud, fstDay, "reviews")
        txt1 += rep
        # time
        (timdata, timsum) = self._splitRepData(
            data,
            (
                (8, colMature, "Mature"),
                (7, colYoung, "Young"),
                (9, colRelearn, "Relearn"),
                (6, colLearn, "Learn"),
                (10, colCram, "Cram"),
            ),
        )
        if self.type == PERIOD_MONTH:
            t = "Minutes"
            convHours = False
        else:
            t = "Hours"
            convHours = True
        txt2 = self._title("Review Time", "The time taken to answer the questions.")
        txt2 += plot("time", timdata, ylabel=t, ylabel2="Cumulative %s" % t)
        rep, tot2 = self._ansInfo(
            timsum, daysStud, fstDay, "minutes", convHours, total=tot
        )
        txt2 += rep
        return self._section(txt1) + self._section(txt2)

    def _ansInfo(
        self,
        totd: list[tuple[int, float]],
        studied: int,
        first: int,
        unit: str,
        convHours: bool = False,
        total: int | None = None,
    ) -> tuple[str, int]:
        assert totd
        tot = totd[-1][1]
        period = self._periodDays()
        if not period:
            # base off earliest repetition date
            period = self._deckAge("review")
        i: list[str] = []
        self._line(
            i,
            "Days studied",
            "<b>%(pct)d%%</b> (%(x)s of %(y)s)"
            % dict(x=studied, y=period, pct=studied / float(period) * 100),
            bold=False,
        )
        if convHours:
            tunit = "hours"
        else:
            tunit = unit
        # T: unit: can be hours, minutes, reviews... tot: the number of unit.
        self._line(i, "Total", "%(tot)s %(unit)s" % dict(unit=tunit, tot=int(tot)))
        if convHours:
            # convert to minutes
            tot *= 60
        self._line(i, "Average for days studied", self._avgDay(tot, studied, unit))
        if studied != period:
            # don't display if you did study every day
            self._line(i, "If you studied every day", self._avgDay(tot, period, unit))
        if total and tot:
            perMin = total / float(tot)
            average_secs = (tot * 60) / total
            self._line(
                i,
                "Average answer time",
                self.col.tr.statistics_average_answer_time(
                    average_seconds=average_secs, cards_per_minute=perMin
                ),
            )
        return self._lineTbl(i), int(tot)

    def _splitRepData(
        self,
        data: list[tuple[Any, ...]],
        spec: Sequence[tuple[int, str, str]],
    ) -> tuple[list[dict[str, Any]], list[tuple[Any, Any]]]:
        sep: dict[int, Any] = {}
        totcnt = {}
        totd: dict[int, Any] = {}
        alltot = []
        allcnt: float = 0
        for n, col, lab in spec:
            totcnt[n] = 0.0
            totd[n] = []
        for row in data:
            for n, col, lab in spec:
                if n not in sep:
                    sep[n] = []
                sep[n].append((row[0], row[n]))
                totcnt[n] += row[n]
                allcnt += row[n]
                totd[n].append((row[0], totcnt[n]))
            alltot.append((row[0], allcnt))
        ret = []
        for n, col, lab in spec:
            if len(totd[n]) and totcnt[n]:
                # bars
                ret.append(dict(data=sep[n], color=col, label=lab))
                # lines
                ret.append(
                    dict(
                        data=totd[n],
                        color=col,
                        label=None,
                        yaxis=2,
                        bars={"show": False},
                        lines=dict(show=True),
                        stack=-n,
                    )
                )
        return (ret, alltot)

    def _added(self, num: int | None = 7, chunk: int = 1) -> Any:
        lims = []
        if num is not None:
            lims.append(
                "id > %d" % ((self.col.sched.day_cutoff - (num * chunk * 86400)) * 1000)
            )
        lims.append("did in %s" % self._limit())
        if lims:
            lim = "where " + " and ".join(lims)
        else:
            lim = ""
        if self.type == PERIOD_MONTH:
            tf = 60.0  # minutes
        else:
            tf = 3600.0  # hours
        return self.col.db.all(
            """
select
(cast((id/1000.0 - ?) / 86400.0 as int))/? as day,
count(id)
from cards %s
group by day order by day"""
            % lim,
            self.col.sched.day_cutoff,
            chunk,
        )

    def _done(self, num: int | None = 7, chunk: int = 1) -> Any:
        lims = []
        if num is not None:
            lims.append(
                "id > %d" % ((self.col.sched.day_cutoff - (num * chunk * 86400)) * 1000)
            )
        lim = self._revlogLimit()
        if lim:
            lims.append(lim)
        if lims:
            lim = "where " + " and ".join(lims)
        else:
            lim = ""
        if self.type == PERIOD_MONTH:
            tf = 60.0  # minutes
        else:
            tf = 3600.0  # hours
        return self.col.db.all(
            f"""
select
(cast((id/1000.0 - ?) / 86400.0 as int))/? as day,
sum(case when type = {REVLOG_LRN} then 1 else 0 end), -- lrn count
sum(case when type = {REVLOG_REV} and lastIvl < 21 then 1 else 0 end), -- yng count
sum(case when type = {REVLOG_REV} and lastIvl >= 21 then 1 else 0 end), -- mtr count
sum(case when type = {REVLOG_RELRN} then 1 else 0 end), -- lapse count
sum(case when type = {REVLOG_CRAM} then 1 else 0 end), -- cram count
sum(case when type = {REVLOG_LRN} then time/1000.0 else 0 end)/?, -- lrn time
-- yng + mtr time
sum(case when type = {REVLOG_REV} and lastIvl < 21 then time/1000.0 else 0 end)/?,
sum(case when type = {REVLOG_REV} and lastIvl >= 21 then time/1000.0 else 0 end)/?,
sum(case when type = {REVLOG_RELRN} then time/1000.0 else 0 end)/?, -- lapse time
sum(case when type = {REVLOG_CRAM} then time/1000.0 else 0 end)/? -- cram time
from revlog %s
group by day order by day"""
            % lim,
            self.col.sched.day_cutoff,
            chunk,
            tf,
            tf,
            tf,
            tf,
            tf,
        )

    def _daysStudied(self) -> Any:
        lims = []
        num = self._periodDays()
        if num:
            lims.append(
                "id > %d" % ((self.col.sched.day_cutoff - (num * 86400)) * 1000)
            )
        rlim = self._revlogLimit()
        if rlim:
            lims.append(rlim)
        if lims:
            lim = "where " + " and ".join(lims)
        else:
            lim = ""
        ret = self.col.db.first(
            """
select count(), abs(min(day)) from (select
(cast((id/1000 - ?) / 86400.0 as int)+1) as day
from revlog %s
group by day order by day)"""
            % lim,
            self.col.sched.day_cutoff,
        )
        assert ret
        return ret

    # Intervals
    ######################################################################

    def ivlGraph(self) -> str:
        (ivls, all, avg, max_), chunk = self._ivls()
        tot = 0
        totd = []
        if not ivls or not all:
            return ""
        for grp, cnt in ivls:
            tot += cnt
            totd.append((grp, tot / float(all) * 100))
        if self.type == PERIOD_MONTH:
            ivlmax = 31
        elif self.type == PERIOD_YEAR:
            ivlmax = 52
        else:
            ivlmax = max(5, ivls[-1][0])
        txt = self._title("Intervals", "Delays until reviews are shown again.")
        txt += self._graph(
            id="ivl",
            ylabel2="Percentage",
            xunit=chunk,
            data=[
                dict(data=ivls, color=colIvl),
                dict(
                    data=totd,
                    color=colCum,
                    yaxis=2,
                    bars={"show": False},
                    lines=dict(show=True),
                    stack=False,
                ),
            ],
            conf=dict(
                xaxis=dict(min=-0.5, max=ivlmax + 0.5),
                yaxes=[dict(), dict(position="right", max=105)],
            ),
        )
        i: list[str] = []
        self._line(i, "Average interval", self.col.format_timespan(avg * 86400))
        self._line(i, "Longest interval", self.col.format_timespan(max_ * 86400))
        return txt + self._lineTbl(i)

    def _ivls(self) -> tuple[list[Any], int]:
        start, end, chunk = self.get_start_end_chunk()
        lim = "and grp <= %d" % end if end else ""
        data = [
            self.col.db.all(
                f"""
select ivl / ? as grp, count() from cards
where did in %s and queue = {QUEUE_TYPE_REV} %s
group by grp
order by grp"""
                % (self._limit(), lim),
                chunk,
            )
        ]
        return (
            data
            + list(
                self.col.db.first(
                    f"""
select count(), avg(ivl), max(ivl) from cards where did in %s and queue = {QUEUE_TYPE_REV}"""
                    % self._limit()
                )
            ),
            chunk,
        )

    # Eases
    ######################################################################

    def easeGraph(self) -> str:
        # 3 + 4 + 4 + spaces on sides and middle = 15
        # yng starts at 1+3+1 = 5
        # mtr starts at 5+4+1 = 10
        d: dict[str, list] = {"lrn": [], "yng": [], "mtr": []}
        types = ("lrn", "yng", "mtr")
        eases = self._eases()
        for type, ease, cnt in eases:
            if type == CARD_TYPE_LRN:
                ease += 5
            elif type == CARD_TYPE_REV:
                ease += 10
            n = types[type]
            d[n].append((ease, cnt))
        ticks = [
            [1, 1],
            [2, 2],
            [3, 3],  # [4,4]
            [6, 1],
            [7, 2],
            [8, 3],
            [9, 4],
            [11, 1],
            [12, 2],
            [13, 3],
            [14, 4],
        ]
        ticks.insert(3, [4, 4])
        txt = self._title(
            "Answer Buttons", "The number of times you have pressed each button."
        )
        txt += self._graph(
            id="ease",
            data=[
                dict(data=d["lrn"], color=colLearn, label="Learning"),
                dict(data=d["yng"], color=colYoung, label="Young"),
                dict(data=d["mtr"], color=colMature, label="Mature"),
            ],
            type="bars",
            conf=dict(xaxis=dict(ticks=ticks, min=0, max=15)),
            ylabel="Answers",
        )
        txt += self._easeInfo(eases)
        return txt

    def _easeInfo(self, eases: list[tuple[int, int, int]]) -> str:
        types = {PERIOD_MONTH: [0, 0], PERIOD_YEAR: [0, 0], PERIOD_LIFE: [0, 0]}
        for type, ease, cnt in eases:
            if ease == 1:
                types[type][0] += cnt
            else:
                types[type][1] += cnt
        i = []
        for type in range(3):
            (bad, good) = types[type]
            tot = bad + good
            try:
                pct = good / float(tot) * 100
            except Exception:
                pct = 0
            i.append(
                "Correct: <b>%(pct)0.2f%%</b><br>(%(good)d of %(tot)d)"
                % dict(pct=pct, good=good, tot=tot)
            )
        return (
            """
<center><table width=%dpx><tr><td width=50></td><td align=center>"""
            % self.width
            + "</td><td align=center>".join(i)
            + "</td></tr></table></center>"
        )

    def _eases(self) -> Any:
        lims = []
        lim = self._revlogLimit()
        if lim:
            lims.append(lim)
        days = self._periodDays()
        if days is not None:
            lims.append(
                "id > %d" % ((self.col.sched.day_cutoff - (days * 86400)) * 1000)
            )
        if lims:
            lim = "and " + " and ".join(lims)
        else:
            lim = ""
        ease4repl = "ease"
        return self.col.db.all(
            f"""
select (case
when type in ({REVLOG_LRN},{REVLOG_RELRN}) then 0
when lastIvl < 21 then 1
else 2 end) as thetype,
(case when type in ({REVLOG_LRN},{REVLOG_RELRN}) and ease = 4 then %s else ease end), count() from revlog where type != {REVLOG_RESCHED} %s
group by thetype, ease
order by thetype, ease"""
            % (ease4repl, lim)
        )

    # Hourly retention
    ######################################################################

    def hourGraph(self) -> str:
        data = self._hourRet()
        if not data:
            return ""
        shifted = []
        counts = []
        mcount = 0
        trend: list[tuple[int, int]] = []
        peak = 0
        for d in data:
            hour = (d[0] - 4) % 24
            pct = d[1]
            if pct > peak:
                peak = pct
            shifted.append((hour, pct))
            counts.append((hour, d[2]))
            if d[2] > mcount:
                mcount = d[2]
        shifted.sort()
        counts.sort()
        if len(counts) < 4:
            return ""
        for d in shifted:
            hour = d[0]
            pct = d[1]
            if not trend:
                trend.append((hour, pct))
            else:
                prev = trend[-1][1]
                diff = pct - prev
                diff /= 3.0
                diff = round(diff, 1)
                trend.append((hour, prev + diff))
        txt = self._title(
            "Hourly Breakdown", "Review success rate for each hour of the day."
        )
        txt += self._graph(
            id="hour",
            data=[
                dict(data=shifted, color=colCum, label="% Correct"),
                dict(
                    data=counts,
                    color=colHour,
                    label="Answers",
                    yaxis=2,
                    bars=dict(barWidth=0.2),
                    stack=False,
                ),
            ],
            conf=dict(
                xaxis=dict(
                    ticks=[
                        [0, "4AM"],
                        [6, "10AM"],
                        [12, "4PM"],
                        [18, "10PM"],
                        [23, "3AM"],
                    ]
                ),
                yaxes=[dict(max=peak), dict(position="right", max=mcount)],
            ),
            ylabel="% Correct",
            ylabel2="Reviews",
        )
        txt += "Hours with less than 30 reviews are not shown."
        return txt

    def _hourRet(self) -> Any:
        lim = self._revlogLimit()
        if lim:
            lim = " and " + lim
        rolloverHour = self.col.conf.get("rollover", 4)
        pd = self._periodDays()
        if pd:
            lim += " and id > %d" % ((self.col.sched.day_cutoff - (86400 * pd)) * 1000)
        return self.col.db.all(
            f"""
select
23 - ((cast((? - id/1000) / 3600.0 as int)) %% 24) as hour,
sum(case when ease = 1 then 0 else 1 end) /
cast(count() as float) * 100,
count()
from revlog where type in ({REVLOG_LRN},{REVLOG_REV},{REVLOG_RELRN}) %s
group by hour having count() > 30 order by hour"""
            % lim,
            self.col.sched.day_cutoff - (rolloverHour * 3600),
        )

    # Cards
    ######################################################################

    def cardGraph(self) -> str:
        # graph data
        div = self._cards()
        d = []
        for c, (t, col) in enumerate(
            (
                ("Mature", colMature),
                ("Young+Learn", colYoung),
                ("Unseen", colUnseen),
                ("Suspended+Buried", colSusp),
            )
        ):
            d.append(dict(data=div[c], label=f"{t}: {div[c]}", color=col))
        # text data
        i: list[str] = []
        (c, f) = self.col.db.first(
            """
select count(id), count(distinct nid) from cards
where did in %s """
            % self._limit()
        )
        self._line(i, "Total cards", c)
        self._line(i, "Total notes", f)
        (low, avg, high) = self._factors()
        if low:
            self._line(i, "Lowest ease", "%d%%" % low)
            self._line(i, "Average ease", "%d%%" % avg)
            self._line(i, "Highest ease", "%d%%" % high)
        info = "<table width=100%>" + "".join(i) + "</table><p>"
        info += """\
A card's <i>ease</i> is the size of the next interval \
when you answer "good" on a review."""
        txt = self._title("Card Types", "The division of cards in your deck(s).")
        txt += "<table width=%d><tr><td>%s</td><td>%s</td></table>" % (
            self.width,
            self._graph(id="cards", data=d, type="pie"),
            info,
        )
        return txt

    def _line(self, i: list[str], a: str, b: int | str, bold: bool = True) -> None:
        # T: Symbols separating first and second column in a statistics table. Eg in "Total:    3 reviews".
        colon = ":"
        if bold:
            i.append(
                ("<tr><td width=200 align=start>%s%s</td><td><b>%s</b></td></tr>")
                % (a, colon, b)
            )
        else:
            i.append(
                ("<tr><td width=200 align=end>%s%s</td><td>%s</td></tr>")
                % (a, colon, b)
            )

    def _lineTbl(self, i: list[str]) -> str:
        return "<table width=400>" + "".join(i) + "</table>"

    def _factors(self) -> Any:
        return self.col.db.first(
            f"""
select
min(factor) / 10.0,
avg(factor) / 10.0,
max(factor) / 10.0
from cards where did in %s and queue = {QUEUE_TYPE_REV}"""
            % self._limit()
        )

    def _cards(self) -> Any:
        return self.col.db.first(
            f"""
select
sum(case when queue={QUEUE_TYPE_REV} and ivl >= 21 then 1 else 0 end), -- mtr
sum(case when queue in ({QUEUE_TYPE_LRN},{QUEUE_TYPE_DAY_LEARN_RELEARN}) or (queue={QUEUE_TYPE_REV} and ivl < 21) then 1 else 0 end), -- yng/lrn
sum(case when queue={QUEUE_TYPE_NEW} then 1 else 0 end), -- new
sum(case when queue<{QUEUE_TYPE_NEW} then 1 else 0 end) -- susp
from cards where did in %s"""
            % self._limit()
        )

    # Footer
    ######################################################################

    def footer(self) -> str:
        b = "<br><br><font size=1>"
        b += "Generated on %s" % time.asctime(time.localtime(time.time()))
        b += "<br>"
        if self.wholeCollection:
            deck = "whole collection"
        else:
            deck = self.col.decks.current()["name"]
        b += "Scope: %s" % deck
        b += "<br>"
        b += "Period: %s" % ["1 month", "1 year", "deck life"][self.type]
        return b

    # Tools
    ######################################################################

    def _graph(
        self,
        id: str,
        data: Any,
        conf: Any | None = None,
        type: str = "bars",
        xunit: int = 1,
        ylabel: str = "Cards",
        ylabel2: str = "",
    ) -> str:
        if conf is None:
            conf = {}
        # display settings
        if type == "pie":
            conf["legend"] = {"container": "#%sLegend" % id, "noColumns": 2}
        else:
            conf["legend"] = {"container": "#%sLegend" % id, "noColumns": 10}
        conf["series"] = dict(stack=True)
        if "yaxis" not in conf:
            conf["yaxis"] = {}
        conf["yaxis"]["labelWidth"] = 40
        if "xaxis" not in conf:
            conf["xaxis"] = {}
        if xunit is None:
            conf["timeTicks"] = False
        else:
            # T: abbreviation of day
            d = "d"
            # T: abbreviation of week
            w = "w"
            # T: abbreviation of month
            mo = "mo"
            conf["timeTicks"] = {1: d, 7: w, 31: mo}[xunit]
        # types
        width = self.width
        height = self.height
        if type == "bars":
            conf["series"]["bars"] = dict(
                show=True, barWidth=0.8, align="center", fill=0.7, lineWidth=0
            )  # pytype: disable=unsupported-operands
        elif type == "barsLine":
            print("deprecated - use 'bars' instead")
            conf["series"]["bars"] = dict(
                show=True, barWidth=0.8, align="center", fill=0.7, lineWidth=3
            )
        elif type == "fill":
            conf["series"]["lines"] = dict(show=True, fill=True)
        elif type == "pie":
            width = int(float(width) / 2.3)
            height = int(float(height) * 1.5)
            ylabel = ""
            conf["series"]["pie"] = dict(
                show=True,
                radius=1,
                stroke=dict(color="#fff", width=5),
                label=dict(
                    show=True,
                    radius=0.8,
                    threshold=0.01,
                    background=dict(opacity=0.5, color="#000"),
                ),
            )
        return """
<table cellpadding=0 cellspacing=10>
<tr>

<td><div style="width: 150px; text-align: center; position:absolute;
 -webkit-transform: rotate(-90deg) translateY(-85px);
font-weight: bold;
">%(ylab)s</div></td>

<td>
<center><div id=%(id)sLegend></div></center>
<div id="%(id)s" style="width:%(w)spx; height:%(h)spx;"></div>
</td>

<td><div style="width: 150px; text-align: center; position:absolute;
 -webkit-transform: rotate(90deg) translateY(65px);
font-weight: bold;
">%(ylab2)s</div></td>

</tr></table>
<script>
$(function () {
    var conf = %(conf)s;
    if (conf.timeTicks) {
        conf.xaxis.tickFormatter = function (val, axis) {
            return val.toFixed(0)+conf.timeTicks;
        }
    }
    conf.yaxis.minTickSize = 1;
    // prevent ticks from having decimals (use whole numbers instead)
    conf.yaxis.tickDecimals = 0;
    conf.yaxis.tickFormatter = function (val, axis) {
            // Just in case we get ticks with decimals, render to one decimal position.  If it's
            // a whole number then render without any decimal (i.e. without the trailing .0).
            return val === Math.round(val) ? val.toFixed(0) : val.toFixed(1);
    }
    if (conf.series.pie) {
        conf.series.pie.label.formatter = function(label, series){
            return '<div class=pielabel>'+Math.round(series.percent)+'%%</div>';
        };
    }
    $.plot($("#%(id)s"), %(data)s, conf);
});
</script>""" % dict(
            id=id,
            w=width,
            h=height,
            ylab=ylabel,
            ylab2=ylabel2,
            data=json.dumps(data),
            conf=json.dumps(conf),
        )

    def _limit(self) -> Any:
        if self.wholeCollection:
            return ids2str([d["id"] for d in self.col.decks.all()])
        return self.col.sched._deck_limit()

    def _revlogLimit(self) -> str:
        if self.wholeCollection:
            return ""
        return "cid in (select id from cards where did in %s)" % ids2str(
            self.col.decks.active()
        )

    def _title(self, title: str, subtitle: str = "") -> str:
        return f"<h1>{title}</h1>{subtitle}"

    def _deckAge(self, by: str) -> int:
        lim = self._revlogLimit()
        if lim:
            lim = " where " + lim
        t = 0
        if by == "review":
            t = self.col.db.scalar("select id from revlog %s order by id limit 1" % lim)
        elif by == "add":
            if self.wholeCollection:
                lim = ""
            else:
                lim = "where did in %s" % ids2str(self.col.decks.active())
            t = self.col.db.scalar("select id from cards %s order by id limit 1" % lim)
        if not t:
            period = 1
        else:
            period = max(1, int(1 + ((self.col.sched.day_cutoff - (t / 1000)) / 86400)))
        return period

    def _periodDays(self) -> int | None:
        start, end, chunk = self.get_start_end_chunk()
        if end is None:
            return None
        return end * chunk

    def _avgDay(self, tot: float, num: int, unit: str) -> str:
        vals = []
        try:
            vals.append("%(a)0.1f %(b)s/day" % dict(a=tot / float(num), b=unit))
            return ", ".join(vals)
        except ZeroDivisionError:
            return ""



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from anki import sync_pb2

# public exports
SyncAuth = sync_pb2.SyncAuth
SyncOutput = sync_pb2.SyncCollectionResponse
SyncStatus = sync_pb2.SyncStatusResponse


# Legacy attributes some add-ons may be using

from .httpclient import HttpClient

AnkiRequestsClient = HttpClient


class Syncer:
    def sync(self) -> str:
        pass



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
from __future__ import annotations

from collections.abc import Callable

import aqt.editor
from anki.collection import OpChanges
from anki.errors import NotFoundError
from aqt import gui_hooks
from aqt.qt import *
from aqt.utils import add_close_shortcut, restoreGeom, saveGeom, tr


class EditCurrent(QMainWindow):
    def __init__(self, mw: aqt.AnkiQt) -> None:
        super().__init__(None, Qt.WindowType.Window)
        self.mw = mw
        self.form = aqt.forms.editcurrent.Ui_Dialog()
        self.form.setupUi(self)
        self.setWindowTitle(tr.editing_edit_current())
        self.setMinimumHeight(400)
        self.setMinimumWidth(250)
        if not is_mac:
            self.setMenuBar(None)
        self.editor = aqt.editor.Editor(
            self.mw,
            self.form.fieldsArea,
            self,
            editor_mode=aqt.editor.EditorMode.EDIT_CURRENT,
        )
        assert self.mw.reviewer.card is not None
        self.editor.card = self.mw.reviewer.card
        self.editor.set_note(self.mw.reviewer.card.note(), focusTo=0)
        restoreGeom(self, "editcurrent")
        close_button = self.form.buttonBox.button(QDialogButtonBox.StandardButton.Close)
        assert close_button is not None
        close_button.setShortcut(QKeySequence("Ctrl+Return"))
        add_close_shortcut(self)
        # qt5.14+ doesn't handle numpad enter on Windows
        self.compat_add_shorcut = QShortcut(QKeySequence("Ctrl+Enter"), self)
        qconnect(self.compat_add_shorcut.activated, close_button.click)
        gui_hooks.operation_did_execute.append(self.on_operation_did_execute)
        self.show()

    def on_operation_did_execute(
        self, changes: OpChanges, handler: object | None
    ) -> None:
        if changes.note_text and handler is not self.editor:
            # reload note
            note = self.editor.note
            try:
                assert note is not None
                note.load()
            except NotFoundError:
                # note's been deleted
                self.cleanup()
                self.close()
                return

            self.editor.set_note(note)

    def cleanup(self) -> None:
        gui_hooks.operation_did_execute.remove(self.on_operation_did_execute)
        self.editor.cleanup()
        saveGeom(self, "editcurrent")
        aqt.dialogs.markClosed("EditCurrent")

    def reopen(self, mw: aqt.AnkiQt) -> None:
        if card := self.mw.reviewer.card:
            self.editor.card = card
            self.editor.set_note(card.note())

    def closeEvent(self, evt: QCloseEvent | None) -> None:
        self.editor.call_after_note_saved(self.cleanup)

    def _saveAndClose(self) -> None:
        self.cleanup()
        self.mw.deferred_delete_and_garbage_collect(self)
        self.close()

    def closeWithCallback(self, onsuccess: Callable[[], None]) -> None:
        def callback() -> None:
            self._saveAndClose()
            onsuccess()

        self.editor.call_after_note_saved(callback)

    onReset = on_operation_did_execute



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
from __future__ import annotations

from typing import cast

from aqt import colors, props
from aqt.qt import *
from aqt.theme import theme_manager


class Switch(QAbstractButton):
    """A horizontal slider to toggle between two states which can be denoted by strings and/or QIcons.

    The left state is the default and corresponds to isChecked()=False.
    The suppoorted slots are toggle(), for an animated transition, and setChecked().
    """

    def __init__(
        self,
        radius: int = 10,
        left_label: str = "",
        right_label: str = "",
        left_color: dict[str, str] | None = None,
        right_color: dict[str, str] | None = None,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent=parent)
        self.setCheckable(True)
        super().setChecked(False)
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self._left_label = left_label
        self._right_label = right_label
        self._left_color = left_color if left_color else colors.ACCENT_CARD
        self._right_color = right_color if right_color else colors.ACCENT_NOTE
        self._path_radius = radius
        self._knob_radius = radius - 2
        self._label_padding = 4
        self._left_knob_position = self._position = radius
        self._right_knob_position = self.width() - self._path_radius
        self._left_label_position = self._label_padding / 2
        self._right_label_position = 2 * self._knob_radius
        self._hide_label: bool = False

    @pyqtProperty(int)  # type: ignore
    def position(self) -> int:
        return self._position

    @position.setter  # type: ignore
    def position(self, position: int) -> None:
        self._position = position
        self.update()

    @property
    def start_position(self) -> int:
        return (
            self._left_knob_position if self.isChecked() else self._right_knob_position
        )

    @property
    def end_position(self) -> int:
        return (
            self._right_knob_position if self.isChecked() else self._left_knob_position
        )

    @property
    def label(self) -> str:
        return self._right_label if self.isChecked() else self._left_label

    @property
    def path_color(self) -> QColor:
        color = self._right_color if self.isChecked() else self._left_color
        return theme_manager.qcolor(color)

    @property
    def label_width(self) -> int:
        font = QFont()
        font.setPixelSize(int(self._knob_radius))
        font.setWeight(QFont.Weight.Bold)
        fm = QFontMetrics(font)
        return (
            max(
                fm.horizontalAdvance(self._left_label),
                fm.horizontalAdvance(self._right_label),
            )
            + 2 * self._label_padding
        )

    def width(self) -> int:
        return self.label_width + 2 * self._path_radius

    def height(self) -> int:
        return 2 * self._path_radius

    def sizeHint(self) -> QSize:
        return QSize(
            self.width(),
            self.height(),
        )

    def setChecked(self, checked: bool) -> None:
        super().setChecked(checked)
        self._position = self.end_position
        self.update()

    def paintEvent(self, _event: QPaintEvent | None) -> None:
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        painter.setPen(Qt.PenStyle.NoPen)
        self._paint_path(painter)
        if not self._hide_label:
            self._paint_label(painter)
        self._paint_knob(painter)

    def _current_path_rectangle(self) -> QRectF:
        return QRectF(
            0,
            0,
            self.width(),
            self.height(),
        )

    def _current_label_rectangle(self) -> QRectF:
        return QRectF(
            (
                self._left_label_position
                if self.isChecked()
                else self._right_label_position
            ),
            0,
            self.label_width,
            self.height(),
        )

    def _current_knob_rectangle(self) -> QRectF:
        return QRectF(
            self.position - self._knob_radius,  # type: ignore
            2,
            2 * self._knob_radius,
            2 * self._knob_radius,
        )

    def _paint_path(self, painter: QPainter) -> None:
        painter.setBrush(QBrush(self.path_color))
        painter.drawRoundedRect(
            self._current_path_rectangle(), self._path_radius, self._path_radius
        )

    def _paint_knob(self, painter: QPainter) -> None:
        color = theme_manager.qcolor(colors.BUTTON_GRADIENT_START)
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(QBrush(color))
        painter.drawEllipse(self._current_knob_rectangle())

    def _paint_label(self, painter: QPainter) -> None:
        painter.setPen(theme_manager.qcolor(colors.CANVAS))
        font = painter.font()
        font.setPixelSize(int(self._knob_radius))
        font.setWeight(QFont.Weight.Bold)
        painter.setFont(font)
        painter.drawText(
            self._current_label_rectangle(), Qt.AlignmentFlag.AlignCenter, self.label
        )

    def mouseReleaseEvent(self, event: QMouseEvent | None) -> None:
        super().mouseReleaseEvent(event)
        assert event is not None
        if event.button() == Qt.MouseButton.LeftButton:
            self._animate_toggle()

    def enterEvent(self, event: QEnterEvent | None) -> None:
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        super().enterEvent(event)

    def toggle(self) -> None:
        super().toggle()
        self._animate_toggle()

    def _animate_toggle(self) -> None:
        animation = QPropertyAnimation(self, cast(QByteArray, b"position"), self)
        animation.setDuration(int(theme_manager.var(props.TRANSITION)))
        animation.setStartValue(self.start_position)
        animation.setEndValue(self.end_position)
        # hide label during animation
        self._hide_label = True
        self.update()

        def on_animation_finished() -> None:
            self._hide_label = False
            self.update()

        qconnect(animation.finished, on_animation_finished)
        # make triggered events execute first so the animation runs smoothly afterwards
        QTimer.singleShot(50, animation.start)
